# from .tools import df_row_to_dict

from .compound import Compound, Ingredient
from .db import Database
from .recipe import Recipe

from numpy import int64, nan, isnan, mean, std
from typing import Callable

import os
import mcol
import mrich


class CompoundTable:
    """Class representing all :class:`.Compound` objects in the 'compound' table of the :class:`.Database`.

    .. attention::

            :class:`.CompoundTable` objects should not be created directly. Instead use the :meth:`.HIPPO.compounds` property. See :doc:`getting_started` and :doc:`insert_elaborations`.

    Use as an iterable
    ==================

    Iterate through :class:`.Compound` objects in the table:

    ::

            for compound in animal.compounds:
                    ...


    Selecting compounds in the table
    ================================

    The :class:`.CompoundTable` can be indexed with :class:`.Compound` IDs, names, aliases, or list/sets/tuples/slices thereof:

    ::

            ctable = animal.compounds

            # indexing individual compounds
            comp = ctable[13]                            # using the ID
            comp = ctable["BSYNRYMUTXBXSQ-UHFFFAOYSA-N"] # using the InChIKey
            comp = ctable["aspirin"]                     # using the alias

            # getting a subset of compounds
            cset = ctable[13,15,18]      # using IDs (tuple)
            cset = ctable[[13,15,18]]    # using IDs (list)
            cset = ctable[set(13,15,18)] # using IDs (set)
            cset = ctable[13:18]         # using a slice

    Tags and base compounds can also be used to filter:

    ::

            cset = animal.compounds(tag='hits') # select compounds tagged with 'hits'
            cset = animal.compounds(base=comp)  # select elaborations of comp

    """

    _table = "compound"
    _name = "all compounds"

    def __init__(
        self,
        db: Database,
    ) -> None:

        self._db = db

    ### FACTORIES

    ### PROPERTIES

    @property
    def db(self) -> Database:
        """Returns the associated :class:`.Database`"""
        return self._db

    @property
    def table(self) -> str:
        """Returns the name of the :class:`.Database` table"""
        return self._table

    @property
    def names(self) -> list[str]:
        """Returns the names of child compounds"""
        result = self.db.select(table=self.table, query="compound_name", multiple=True)
        return [q for q, in result]

    @property
    def name(self):
        return self._name

    @property
    def ids(self) -> list[int]:
        """Returns the IDs of child compounds"""
        result = self.db.select(table=self.table, query="compound_id", multiple=True)
        return [q for q, in result]

    @property
    def str_ids(self) -> str:
        """Return an SQL formatted tuple string of the :class:`.Compound` IDs"""
        return str(tuple(self.ids)).replace(",)", ")")

    @property
    def inchikeys(self) -> list[str]:
        """Returns the inchikeys of all compounds"""
        result = self.db.select(
            query="compound_inchikey",
            table="compound",
            multiple=True,
        )
        return [q for q, in result]

    @property
    def tags(self) -> set[str]:
        """Returns the set of unique tags present in this compound set"""
        values = self.db.select_where(
            table="tag",
            query="DISTINCT tag_name",
            key="tag_compound IS NOT NULL",
            multiple=True,
        )
        return set(v for v, in values)

    @property
    def reactants(self) -> "CompoundSet":
        """Returns a :class:`.CompoundSet` of all compounds that are used as a reactants"""
        # ids = self.db.select(table='reactant', query='DISTINCT reactant_compound', multiple=True)
        ids = self.db.execute(
            "SELECT reactant_compound FROM reactant LEFT JOIN reaction ON reactant.reactant_compound = reaction.reaction_product WHERE reaction.reaction_product IS NULL"
        ).fetchall()
        ids = [q for q, in ids]
        from .cset import CompoundSet

        cset = CompoundSet(self.db, ids)
        cset._name = "all reactants"
        return cset

    @property
    def products(self) -> "CompoundSet":
        """Returns a :class:`.CompoundSet` of all compounds that are a product of a reaction but not a reactant"""
        ids = self.db.execute(
            "SELECT reaction_product FROM reaction LEFT JOIN reactant ON reaction.reaction_product = reactant.reactant_compound WHERE reactant.reactant_compound IS NULL"
        ).fetchall()
        ids = [q for q, in ids]
        from .cset import CompoundSet

        cset = CompoundSet(self.db, ids)
        cset._name = "all products"
        return cset

    @property
    def intermediates(self) -> "CompoundSet":
        """Returns a :class:`.CompoundSet` of all compounds that are products and reactants"""
        ids = self.db.execute(
            "SELECT DISTINCT reaction_product FROM reaction INNER JOIN reactant ON reaction.reaction_product = reactant.reactant_compound"
        ).fetchall()
        ids = [q for q, in ids]
        from .cset import CompoundSet

        cset = CompoundSet(self.db, ids)
        cset._name = "all intermediates"
        return cset

    @property
    def num_reactants(self) -> int:
        """Returns the number of reactants (see :meth:`CompoundTable.reactants`)"""
        return len(self.reactants)

    @property
    def num_intermediates(self) -> int:
        """Returns the number of intermediates (see :meth:`CompoundTable.intermediates`)"""
        return len(self.intermediates)

    @property
    def num_products(self) -> int:
        """Returns the number of products (see :meth:`CompoundTable.products`)"""
        return len(self.products)

    @property
    def elabs(self) -> "CompoundSet":
        """Returns a :class:`.CompoundSet` of all compounds that are a an elaboration of an existing base"""
        ids = self.db.select_where(
            query="scaffold_superstructure",
            table="scaffold",
            key="scaffold_superstructure IS NOT NULL",
            multiple=True,
            none="quiet",
        )

        if not ids:
            return None

        ids = [q for q, in ids]
        from .cset import CompoundSet

        cset = CompoundSet(self.db, ids)
        cset._name = "all elaborations"
        return cset

    @property
    def bases(self) -> "CompoundSet":
        """Returns a :class:`.CompoundSet` of all compounds that are the basis for a set of elaborations"""
        ids = self.db.select_where(
            query="DISTINCT scaffold_base",
            table="scaffold",
            key="scaffold_base IS NOT NULL",
            multiple=True,
            none="quiet",
        )
        ids = [q for q, in ids]
        from .cset import CompoundSet

        cset = CompoundSet(self.db, ids)
        cset._name = "all bases"
        return cset

    @property
    def num_elabs(self) -> int:
        """Returns the number of compounds that are a an elaboration of an existing base"""
        return len(self.elabs)

    @property
    def num_bases(self) -> int:
        """Returns the number of compounds that are the basis for a set of elaborations"""
        return len(self.bases)

    ### METHODS

    def get_by_tag(
        self,
        tag: str,
        inverse: bool = False,
    ) -> "CompoundSet":
        """Get all child compounds with a certain tag

        :param tag: tag to filter by

        """

        if not inverse:

            values = self.db.select_where(
                query="tag_compound", table="tag", key="name", value=tag, multiple=True
            )

        else:

            values = self.db.select_where(
                query="tag_compound", table="tag", key="name", value=tag, multiple=True
            )

            if not values:
                return self

            ids = [v for v, in values if v]

            values = self.db.select_where(
                query="compound_id",
                table="compound",
                key=f"compound_id NOT IN {str(tuple(ids))}",
                multiple=True,
            )

        if not values:
            return None

        ids = [v for v, in values if v]
        cset = self[ids]
        if inverse:
            cset._name = f"compounds not tagged {tag}"
        else:
            cset._name = f"compounds tagged {tag}"
        return cset

    def get_by_metadata(
        self,
        key: str,
        value: str | None = None,
    ):
        """Get all child compounds by their metadata. If no value is passed, then simply containing the key in the metadata dictionary is sufficient

        :param key: metadata key
        :param value: metadata value (Default value = None)

        """
        results = self.db.select(
            query="compound_id, compound_metadata", table="compound", multiple=True
        )
        if value is None:
            ids = [i for i, d in results if d and f'"{key}":' in d]
            name = f"compounds with {key} in metadata"
        else:
            if isinstance(value, str):
                value = f'"{value}"'
            ids = [i for i, d in results if d and f'"{key}": {value}' in d]
            name = f"compounds with metadata[{key}] == {value}"

        cset = self[ids]
        cset._name = name
        return cset

    def get_by_base(
        self,
        base: Compound | int,
    ) -> "CompoundSet":
        """Get all compounds that elaborate the given base compound

        :param base: :class:`.Compound` object or ID to search by

        """

        if not isinstance(base, int):
            assert base._table == "compound"
            base = base.id

        values = self.db.select_where(
            query="scaffold_superstructure",
            table="scaffold",
            key="base",
            value=base,
            multiple=True,
        )
        ids = [v for v, in values if v]
        cset = self[ids]
        cset._name = f"elaborations of C{base}"
        return cset

    def get_by_smiles(self, smiles: str, **kwargs) -> "Compound | None":
        """Get a member compound by its smiles"""

        from .tools import inchikey_from_smiles, sanitise_smiles, SanitisationError

        assert isinstance(smiles, str), f"Non-string {smiles=}"
        try:
            smiles = sanitise_smiles(smiles, sanitisation_failed="error")
        except SanitisationError as e:
            mrich.error(f"Could not sanitise {smiles=}")
            mrich.error(str(e))
            return None
        except AssertionError:
            mrich.error(f"Could not sanitise {smiles=}")
            return None
            return c
        inchikey = inchikey_from_smiles(smiles)
        return self.db.get_compound(inchikey=inchikey, **kwargs)

    def summary(self) -> None:
        """Print a summary of this compound set"""
        mrich.header("CompoundTable()")
        mrich.var("#compounds", len(self))
        # mrich.var('#poses', self.num_poses)
        mrich.var("tags", self.tags)
        mrich.var("#bases", self.num_bases)
        mrich.var("#elabs", self.num_elabs)
        mrich.var("#reactants", self.num_reactants)
        mrich.var("#intermediates", self.num_intermediates)
        mrich.var("#products", self.num_products)

    def draw(self) -> None:
        """2D grid of drawings of molecules in this set

        .. attention::

                This method instantiates a :class:`.CompoundSet` containing all compounds, it is recommended to instead select a subset for display. This method is only intended for use within a Jupyter Notebook.

        """
        return self[self.ids].draw()

    def interactive(self) -> None:
        """Interactive widget to navigate compounds in the table

        .. attention::

                This method instantiates a :class:`.CompoundSet` containing all compounds, it is recommended to instead select a subset for display. This method is only intended for use within a Jupyter Notebook.

        """
        self[self.ids].interactive()

    def plot_tsnee(self, **kwargs) -> "go.Figure":
        """Plot a tanimoto similarity plot of these compounds"""
        return self[:].plot_tsnee(**kwargs)

    def write_smiles_csv(self, file: str) -> None:
        """Write a CSV of the smiles contained in this set to a file

        :param file: path of the CSV file

        """
        from pandas import DataFrame

        records = self.db.execute(
            """SELECT compound_id, compound_smiles FROM compound ORDER BY compound_id"""
        ).fetchall()

        data = [dict(id=id, smiles=smiles) for id, smiles in records]

        df = DataFrame(data)
        mrich.writing(file)
        df.to_csv(file, index=False)

    ### DUNDERS

    def __call__(
        self,
        *,
        tag: str = None,
        base: int | Compound = None,
        smiles: str | None = None,
        **kwargs,
    ) -> "CompoundSet | Compound | None":
        """Filter compounds by a given tag, base, or it's SMILES string. See :meth:`.CompoundTable.get_by_tag` and :meth:`.CompoundTable.get_by_base`

        :param tag: optional tag to filter by
        :param base: optional :class:`.Compound` ID or object to filter by
        :param base: optional SMILES string to filter by
        :returns: :class:`.CompoundSet` if searching by tag or base, else :class:`.Compound` object

        """

        if tag:
            return self.get_by_tag(tag, **kwargs)
        elif base:
            return self.get_by_base(base, **kwargs)
        elif smiles:
            return self.get_by_smiles(smiles, **kwargs)
        else:
            mrich.error("Must provide one of tag, base, or smiles arguments")
            return None

    def __getitem__(
        self,
        key: int | str | tuple | list | set | slice,
    ) -> Compound:
        """Get a member :class:`.Pose` object or subset :class:`.PoseSet` thereof.

        :param key: Can be an integer ID, negative integer index, alias or inchikey string, list/set/tuple of IDs, or slice of IDs

        """

        match key:

            # case int():
            case key if isinstance(key, int) or isinstance(key, int64):

                if key == 0:
                    return self.__getitem__(key=1)

                if key < 0:
                    key = len(self) + 1 + key
                    return self.__getitem__(key=key)

                else:
                    return self.db.get_compound(id=key)

            case str():
                comp = self.db.get_compound(inchikey=key, none="quiet")
                if not comp:
                    comp = self.db.get_compound(alias=key)
                return comp

            case key if (
                isinstance(key, list) or isinstance(key, tuple) or isinstance(key, set)
            ):

                indices = []
                for i in key:
                    if isinstance(i, int) or isinstance(i, int64):
                        index = i
                    elif isinstance(i, float):
                        index = int(i)
                    elif isinstance(i, str):
                        index = self.db.get_compound_id(inchikey=i)
                    else:
                        raise NotImplementedError

                    assert index
                    indices.append(index)

                return CompoundSet(self.db, indices)

            case slice():
                ids = self.db.slice_ids(
                    table=self.table, start=key.start, stop=key.stop, step=key.step
                )
                return self[ids]

            case _:
                mrich.error(
                    f"Unsupported type for CompoundTable.__getitem__(): {key=} {type(key)}"
                )

        return None

    def __str__(self) -> str:
        """Unformatted string representation"""

        if self.name:
            s = f"{self.name}: "
        else:
            s = ""

        s += "{" f"C × {len(self)}" "}"

        return s

    def __repr__(self) -> str:
        """ANSI ormatted string representation"""
        return f"{mcol.bold}{mcol.underline}{self}{mcol.unbold}{mcol.ununderline}"

    def __rich__(self) -> str:
        """Representation for mrich"""
        return f"[bold underline]{self}"

    def __len__(self) -> int:
        """Total number of compounds"""
        return self.db.count(self.table)

    def __iter__(self):
        """Iterate through all compounds"""
        return iter(self[i + 1] for i in range(len(self)))


class CompoundSet:
    """Object representing a subset of the 'compound' table in the :class:`.Database`.

    .. attention::

            :class:`.CompoundSet` objects should not be created directly. Instead use the :meth:`.HIPPO.compounds` property. See :doc:`getting_started` and :doc:`insert_elaborations`.

    Use as an iterable
    ==================

    Iterate through :class:`.Compound` objects in the set:

    ::

            cset = animal.compounds[:100]

            for compound in cset:
                    ...

    Check membership
    ================

    To determine if a :class:`.Compound` is present in the set:

    ::

            is_member = compound in cset

    Selecting compounds in the set
    ==============================

    The :class:`.CompoundSet` can be indexed like standard Python lists by their indices

    ::

            cset = animal.compounds[1:100]

            # indexing individual compounds
            comp = cset[0]  # get the first compound
            comp = cset[1]  # get the second compound
            comp = cset[-1] # get the last compound

            # getting a subset of compounds using a slice
            cset2 = cset[13:18] # using a slice

    Tags and base compounds can also be used to filter:

    ::

            cset = animal.compounds(tag='hits') # select compounds tagged with 'hits'
            cset = animal.compounds(base=comp)  # select elaborations of comp

    """

    _table = "compound"

    def __init__(
        self,
        db: Database,
        indices: list = None,
        sort: bool = True,
        name: str | None = None,
    ):

        self._db = db

        indices = indices or []

        if not isinstance(indices, list):
            indices = list(indices)

        indices = [int(i) for i in indices]

        if sort:
            self._indices = sorted(list(set(indices)))
        else:
            self._indices = list(set(indices))

        self._name = name
        self._total_changes = db.total_changes

    ### PROPERTIES

    @property
    def db(self) -> "Database":
        """ """
        return self._db

    @property
    def table(self) -> str:
        """Get the name of the database table"""
        return self._table

    @property
    def indices(self) -> list[int]:
        """Returns the ids of compounds in this set"""
        return self._indices

    @property
    def ids(self) -> list[int]:
        """Returns the ids of compounds in this set"""
        return self.indices

    @property
    def name(self) -> str | None:
        """Returns the name of set"""
        return self._name

    @property
    def names(self) -> list[str]:
        """Returns the aliases of compounds in this set"""
        result = self.db.select_where(
            query="compound_alias",
            table="compound",
            key=f"compound_id in {self.str_ids}",
            multiple=True,
        )
        return [q for q, in result]

    @property
    def smiles(self) -> list[str]:
        """Returns the smiles of child compounds"""
        result = self.db.select_where(
            query="compound_smiles",
            table="compound",
            key=f"compound_id in {self.str_ids}",
            multiple=True,
        )
        return [q for q, in result]

    @property
    def mols(self) -> "list[Chem.Mol]":
        """Returns the molecules of child compounds"""
        from rdkit.Chem import Mol

        result = self.db.select_where(
            query="mol_to_binary_mol(compound_mol)",
            table="compound",
            key=f"compound_id in {self.str_ids}",
            multiple=True,
        )
        return [Mol(q) for q, in result]

    @property
    def inchikeys(self) -> list[str]:
        """Returns the inchikeys of compounds in this set"""
        result = self.db.select_where(
            query="compound_inchikey",
            table="compound",
            key=f"compound_id in {self.str_ids}",
            multiple=True,
        )
        return [q for q, in result]

    @property
    def tags(self) -> set[str]:
        """Returns the set of unique tags present in this compound set"""
        values = self.db.select_where(
            table="tag",
            query="DISTINCT tag_name",
            key=f"tag_compound in {self.str_ids}",
            multiple=True,
        )
        if not values:
            return set()
        return set(v for v, in values)

    @property
    def num_poses(self) -> int:
        """Count the poses associated to this set of compounds"""
        from .pset import PoseSet

        return self.db.count_where(table="pose", key=f"pose_compound in {self.str_ids}")

    @property
    def poses(self) -> "PoseSet":
        """Get the poses associated to this set of compounds"""
        from .pset import PoseSet

        ids = self.db.select_where(
            query="pose_id",
            table="pose",
            key=f"pose_compound in {self.str_ids}",
            multiple=True,
            none="warning",
        )

        if not ids:
            return PoseSet(self.db, {})

        ids = [v for v, in ids]
        return PoseSet(self.db, ids)

    @property
    def best_placed_poses(self) -> "PoseSet":
        """Get the best placed pose for each compound in this set"""
        from .pset import PoseSet

        query = self.db.select_where(
            table="pose",
            query="pose_id, MIN(pose_distance_score)",
            key=f"pose_compound in {self.str_ids} GROUP BY pose_compound",
            multiple=True,
        )
        ids = [i for i, s in query]
        return PoseSet(self.db, ids)

    @property
    def str_ids(self) -> str:
        """Return an SQL formatted tuple string of the :class:`.Compound` IDs"""
        return str(tuple(self.ids)).replace(",)", ")")

    @property
    def num_heavy_atoms(self) -> int:
        """Get the total number of heavy atoms"""
        return sum([c.num_heavy_atoms for c in self])

    @property
    def num_rings(self):
        """Get the total number of molecular rings"""
        return sum([c.num_rings for c in self])

    @property
    def formula(self) -> str:
        """Get the combined chemical formula for all compounds"""
        from molparse.atomtypes import atomtype_dict_to_formula

        return atomtype_dict_to_formula(self.atomtype_dict)

    @property
    def atomtype_dict(self) -> dict[str, int]:
        """Get a dictionary with atomtypes as keys and corresponding quantities/counts as values"""
        from molparse.atomtypes import formula_to_atomtype_dict, combine_atomtype_dicts

        atomtype_dicts = [c.atomtype_dict for c in self]
        return combine_atomtype_dicts(atomtype_dicts)

    @property
    def num_atoms_added(self) -> list[int]:
        """Calculate the number of atoms added w.r.t the base

        :returns: list of number of atoms added values

        """

        query = self.db.execute(
            f"""
        WITH nums AS (
            SELECT A.compound_id AS comp_id, 
            mol_num_hvyatms(A.compound_mol) - mol_num_hvyatms(B.compound_mol) AS diff 
            FROM compound A, compound B
            WHERE A.compound_base = B.compound_id
            AND A.compound_id IN {self.str_ids}
        )

        SELECT compound_id, diff FROM compound
        LEFT JOIN nums
        ON comp_id = compound_id
        WHERE compound_id IN {self.str_ids}
        """
        ).fetchall()

        lookup = {k: v for k, v in query}

        return [lookup[i] for i in self.indices]

    @property
    def avg_num_atoms_added(self) -> float:
        """Calculate the average number of atoms added w.r.t the base

        :returns: average number of atoms added values for compounds which have a base

        """

        (avg,) = self.db.execute(
            f"""
        WITH nums AS (
            SELECT A.compound_id AS comp_id, 
            mol_num_hvyatms(A.compound_mol) - mol_num_hvyatms(B.compound_mol) AS diff 
            FROM compound A, compound B
            WHERE A.compound_base = B.compound_id
            AND A.compound_id IN {self.str_ids}
        )

        SELECT AVG(diff) FROM compound
        INNER JOIN nums
        ON comp_id = compound_id
        WHERE compound_id IN {self.str_ids}
        """
        ).fetchone()

        return avg

    @property
    def risk_diversity(self) -> float:
        """Calculate the average spread of risk (#atoms added) for each base in this set

        :returns: average of the standard deviations of number of atoms added for each base

        """

        return self.get_risk_diversity()

    @property
    def elaboration_balance(self) -> float:
        """Measure of how evenly elaborations are distributed across bases in this set"""

        sql = f"""
        SELECT COUNT(1) FROM scaffold
        WHERE scaffold_superstructure IN {self.str_ids}
        GROUP BY scaffold_base
        """

        counts = self.db.execute(sql).fetchall()

        counts = [c for c, in counts]  # + [0 for _ in range(len(self)-len(counts))]

        from hirsch import hirsch

        return hirsch(counts)

        # return -std(counts)

    @property
    def num_bases_elaborated(self) -> int:
        """Count the number of base compounds that have at least one elaboration in this set

        :returns: number of base compounds

        """

        (count,) = self.db.execute(
            f"""
                SELECT COUNT(DISTINCT scaffold_base) FROM scaffold
                WHERE scaffold_superstructure IN {self.str_ids}  
            """
        ).fetchone()

        return count

    @property
    def bases(self) -> "CompoundSet":
        """Get the base compounds that have at least one elaboration in this set

        :returns: :class:`.CompoundSet`

        """

        base_ids = self.db.execute(
            f"""
                SELECT DISTINCT scaffold_base FROM scaffold
                WHERE scaffold_superstructure IN {self.str_ids}  
            """
        ).fetchall()

        base_ids = [i for i, in base_ids]
        return CompoundSet(self.db, base_ids)

    @property
    def elabs(self) -> "CompoundSet":
        """Returns a :class:`.CompoundSet` of all compounds that are a an elaboration of an existing base"""

        ids = self.db.select_where(
            query="scaffold_superstructure",
            table="scaffold",
            key=f"scaffold_superstructure IS NOT NULL and scaffold_base IN {self.str_ids}",
            multiple=True,
            none="quiet",
        )

        if not ids:
            return None

        ids = [q for q, in ids]
        from .cset import CompoundSet

        return CompoundSet(self.db, ids)

    @property
    def elab_df(self) -> "pd.DataFrame":
        """Get a DataFrame summarising the elaborations in this CompoundSet"""
        from pandas import DataFrame

        cluster_dict = self.db.get_compound_cluster_dict(max_bases=1)

        data = []
        for scaffold, elabs in cluster_dict.items():
            scaffold = self.db.get_compound(id=scaffold[0])
            elabs = CompoundSet(self.db, indices=elabs)
            data.append(
                dict(
                    scaffold_id=scaffold.id,
                    scaffold_compound=scaffold,
                    elabs=elabs,
                    num_elabs=len(elabs),
                )
            )

        return DataFrame(data)

    @property
    def id_num_poses_dict(self) -> dict[int, int]:
        """Get a dictionary mapping compound ids to the number of poses"""

        sql = f"""
            SELECT pose_compound, COUNT(1) FROM pose
            WHERE pose_compound IN {self.str_ids}
            GROUP BY pose_compound
        """

        records = self.db.execute(sql)

        assert records

        lookup = {k: v for k, v in records}

        for id in self.ids:
            if id not in lookup:
                lookup[id] = 0

        return lookup

    @property
    def _db_changed(self) -> bool:
        """Has the database changed?"""
        if self._total_changes != self.db.total_changes:
            self._total_changes = self.db.total_changes
            return True
        return False

    @property
    def reaction_ids(self) -> list[int]:
        """Returns a list of :class:`.Reaction` IDs that result in members of this set"""
        records = self.db.select_where(
            table="reaction",
            query="reaction_id",
            key=f"reaction_product IN {self.str_ids}",
            multiple=True,
        )
        if not records:
            return None
        return [r for r, in records]

    ### FILTERING

    def get_by_tag(
        self,
        tag: str,
        inverse: bool = False,
    ) -> "CompoundSet":
        """Get all child compounds with a certain tag"""

        values = self.db.select_where(
            query="tag_compound",
            table="tag",
            key=f'tag_name = "{tag}" AND tag_compound IN {self.str_ids}',
            multiple=True,
        )

        if inverse:
            matches = set(v for v, in values)
            ids = [i for i in self.ids if i not in matches]
        else:
            ids = [v for v, in values]

        return CompoundSet(self.db, ids)

    def get_by_metadata(self, key: str, value: str | None = None) -> "CompoundSet":
        """Get all child compounds with by their metadata. If no value is passed, then simply containing the key in the metadata dictionary is sufficient

        :param key: metadata key
        :param value: metadata value (Default value = None)
        """

        results = self.db.select(
            query="compound_id, compound_metadata", table="compound", multiple=True
        )
        if value is None:
            ids = [i for i, d in results if d and f'"{key}":' in d and i in self.ids]
        else:
            if isinstance(value, str):
                value = f'"{value}"'
            ids = [
                i
                for i, d in results
                if d and f'"{key}": {value}' in d and i in self.ids
            ]
        return CompoundSet(self.db, ids)

    def get_by_base(
        self,
        base: Compound | int,
        none: str = "error",
    ) -> "CompoundSet":
        """Get all compounds that elaborate the given base compound

        :param base: :class:`.Compound` object or ID to search by

        """

        if not isinstance(base, int):
            assert base._table == "compound"
            base = base.id

        values = self.db.select_where(
            query="scaffold_superstructure",
            table="scaffold",
            key=f"scaffold_base = {base} AND scaffold_superstructure IN {self.str_ids}",
            multiple=True,
            none=none,
        )
        ids = [v for v, in values if v]

        if not ids:
            return None
        return CompoundSet(self.db, ids)

    def get_all_possible_reactants(
        self,
        debug: bool = False,
    ) -> "CompoundSet":
        """Recursively searches for all the reactants that could possible be needed to synthesise these compounds.

        :param debug: Increased verbosity for debugging (Default value = False)

        """
        all_reactants, all_reactions = self.db.get_unsolved_reaction_tree(
            product_ids=self.ids, debug=debug
        )
        return all_reactants

    def get_all_possible_reactions(
        self,
        debug: bool = False,
    ) -> "ReactionSet":
        """Recursively searches for all the reactants that could possible be needed to synthesise these compounds.

        :param debug: Increased verbosity for debugging (Default value = False)

        """
        all_reactants, all_reactions = self.db.get_unsolved_reaction_tree(
            product_ids=self.ids, debug=debug
        )
        return all_reactions

    def get_risk_diversity(self, debug: bool = False) -> float:
        """Calculate the average spread of risk (#atoms added) for each base in this set

        :returns: average of the standard deviations of number of atoms added for each base

        """

        variances = self.db.execute(
            f"""
        WITH nums AS (
            SELECT scaffold_base AS base, scaffold_superstructure AS elab, 
            mol_num_hvyatms(c2.compound_mol) - mol_num_hvyatms(c1.compound_mol) AS diff
            FROM scaffold
            INNER JOIN compound AS c1 ON scaffold_base = c1.compound_id
            INNER JOIN compound AS c2 ON scaffold_superstructure = c2.compound_id
            WHERE scaffold_superstructure IN {self.str_ids}
        ),

        means AS (  
            SELECT base, AVG(diff) AS mean FROM nums
            GROUP BY base
        )

        SELECT AVG((nums.diff - mean)*(nums.diff - mean)) var FROM nums
        LEFT JOIN means
        ON nums.base = means.base
        GROUP BY nums.base
        """
        ).fetchall()

        if not variances:
            return None

        variances = [v for v, in variances]

        if debug:
            mrich.debug(f"{variances=}")

        return mean(variances)

    def count_by_tag(
        self,
        tag: str,
    ) -> "CompoundSet":
        """Count all child compounds with a certain tag

        :param tag: tag to filter by

        """
        (count,) = self.db.select_where(
            query="COUNT(tag_compound)",
            table="tag",
            key=f'tag_name = "{tag}" AND tag_compound IN {self.str_ids}',
            multiple=False,
        )
        return count

    ### CONSOLE / NOTEBOOK OUTPUT

    def draw(self) -> None:
        """Draw a grid of all contained molecules.

        .. attention::

                This method is only intended for use within a Jupyter Notebook.

        """

        from molparse.rdkit import draw_grid

        data = [(str(c), c.mol) for c in self]

        mols = [d[1] for d in data]
        labels = [d[0] for d in data]

        display(draw_grid(mols, labels=labels))

    def grid(self) -> None:
        """Draw a grid of all contained molecules.

        .. attention::

                This method is only intended for use within a Jupyter Notebook.

        """

        self.draw()

    def summary(self) -> None:
        """Print a summary of this compound set"""
        mrich.header("CompoundSet()")
        mrich.var("#compounds", len(self))
        mrich.var("#poses", self.num_poses)
        mrich.var("tags", self.tags)

    def interactive(
        self,
        function: Callable | None = None,
    ) -> None:
        """Creates a ipywidget to interactively navigate this PoseSet."""

        from ipywidgets import (
            interactive,
            BoundedIntText,
            Checkbox,
            interactive_output,
            HBox,
            GridBox,
            Layout,
            VBox,
        )
        from IPython.display import display
        from pprint import pprint

        if function:

            def widget(i):
                compound = self[i]
                display(compound)
                function(compound)

            return interactive(
                widget,
                i=BoundedIntText(
                    value=0,
                    min=0,
                    max=len(self) - 1,
                    step=1,
                    description=f"Comp (/{len(self)}):",
                    disabled=False,
                ),
            )

        else:

            a = BoundedIntText(
                value=0,
                min=0,
                max=len(self) - 1,
                step=1,
                description=f"Comp (/{len(self)}):",
                disabled=False,
            )

            b = Checkbox(description="Name", value=True)
            c = Checkbox(description="Summary", value=False)
            d = Checkbox(description="2D", value=True)
            e = Checkbox(description="Poses", value=False)
            f = Checkbox(description="Reactions", value=False)
            g = Checkbox(description="Tags", value=False)
            h = Checkbox(description="Quotes", value=False)
            i = Checkbox(description="Metadata", value=False)
            j = Checkbox(description="Classify", value=False)

            ui1 = GridBox(
                [b, c, d], layout=Layout(grid_template_columns="repeat(3, 100px)")
            )
            ui2 = GridBox(
                [e, f, g], layout=Layout(grid_template_columns="repeat(3, 100px)")
            )
            ui3 = GridBox(
                [h, i, j], layout=Layout(grid_template_columns="repeat(3, 100px)")
            )
            ui = VBox([a, ui1, ui2, ui3])

            def widget(
                i,
                name=True,
                summary=True,
                draw=True,
                poses=True,
                reactions=True,
                tags=True,
                quotes=True,
                metadata=True,
                classify=True,
            ):
                """

                :param i: param name:  (Default value = True)
                :param summary: Default value = True)
                :param draw: Default value = True)
                :param poses: Default value = True)
                :param reactions: Default value = True)
                :param metadata: Default value = True)
                :param name:  (Default value = True)

                """
                comp = self[i]

                if name and not summary:
                    print(repr(comp))

                if summary:
                    comp.summary(metadata=False, draw=False, tags=False)

                if draw:
                    comp.draw()

                if poses and (pset := comp.poses):
                    for p in pset:
                        mrich.print(p)
                    pset.draw()

                if reactions and (reactions := comp.reactions):
                    for r in reactions:
                        mrich.print(r)
                        r.draw()

                if tags:
                    mrich.title("Tags")
                    mrich.print(comp.tags)

                if quotes:
                    mrich.title("Quotes")
                    display(comp.get_quotes(df=True))

                if metadata:
                    mrich.title("Metadata:")
                    mrich.print(comp.metadata)

                if classify:
                    mrich.title("Classification:")
                    comp.classify()

            out = interactive_output(
                widget,
                {
                    "i": a,
                    "name": b,
                    "summary": c,
                    "draw": d,
                    "poses": e,
                    "reactions": f,
                    "tags": g,
                    "quotes": h,
                    "metadata": i,
                    "classify": j,
                },
            )

            display(ui, out)

    ### OTHER METHODS

    def add(self, compound: Compound | int) -> None:
        """Add a compound to this set

        :param compound: compound to be added

        """

        if isinstance(compound, Compound):
            compound = compound.id

        if compound not in self.ids:
            from bisect import insort

            insort(self.ids, compound)

    def get_recipes(
        self,
        amount: float = 1,
        debug: bool = False,
        pick_cheapest: bool = False,
        permitted_reactions: "ReactionSet | None" = None,
        quoted_only: bool = False,
        supplier: None | str = None,
        **kwargs,
    ):
        """Generate the :class:`.Recipe` to make these compounds.

        See :meth:`.Recipe.from_compounds`
        """
        from .recipe import Recipe

        return Recipe.from_compounds(
            self,
            amount=amount,
            debug=debug,
            pick_cheapest=pick_cheapest,
            permitted_reactions=permitted_reactions,
            quoted_only=quoted_only,
            supplier=supplier,
            **kwargs,
        )

    def get_routes(
        self,
        permitted_reactions: "None | ReactionSet" = None,
        debug: bool = True,
    ) -> "RouteSet":
        """Get a RoutSet to products in this set.

        :param permitted_reactions: optionally restrict reactions to those in this :class:`.ReactionSet`

        """

        if "route" not in self.db.table_names:
            mrich.error("route table not in Database")
            raise NotImplementedError

        if permitted_reactions is not None:

            sql = f"""
            SELECT route_id, route_product, component_ref FROM route
            INNER JOIN component ON route_id = component_route
            WHERE route_product IN {self.str_ids}
            AND component_type = 1
            """

            permitted_reactions = set(permitted_reactions.ids)

            if debug:
                mrich.debug("Querying database for routes")
            records = self.db.execute(sql).fetchall()

            if debug:
                mrich.debug("Assembling route dictionary")

            routes = {}
            for route_id, route_product, reaction_id in records:
                if route_id not in routes:
                    routes[route_id] = dict(product=route_product, reactions=set())
                assert routes[route_id]["product"] == route_product
                routes[route_id]["reactions"].add(reaction_id)

            if debug:
                mrich.debug("Checking availability")

            available_routes = set()
            for route_id, route_dict in routes.items():
                product = route_dict["product"]
                assert product in self
                reactions = route_dict["reactions"]
                if all(r in permitted_reactions for r in reactions):
                    available_routes.add(route_id)

            routes = [
                self.db.get_route(id=route_id)
                for route_id in mrich.track(available_routes, prefix="Getting routes")
            ]

        else:

            sql = f"""
            SELECT route_id FROM route
            WHERE route_product IN {self.str_ids}
            """

            if debug:
                mrich.debug("Querying database for routes")
            records = self.db.execute(sql).fetchall()

            routes = [
                self.db.get_route(id=route_id)
                for route_id, in mrich.track(records, prefix="Getting routes")
            ]

        from .recipe import RouteSet

        return RouteSet(self.db, routes)

    def copy(self) -> "CompoundSet":
        """Returns a copy of this set"""
        return CompoundSet(self.db, self.ids)

    def shuffled(self) -> "CompoundSet":
        """Returns a randomised copy of this set"""
        copy = self.copy()
        copy.shuffle()
        return copy

    def pop(self) -> Compound:
        """Pop the last compound in this set"""
        c_id = self.pop_id()
        return self.db.get_compound(id=c_id)

    def pop_id(self) -> int:
        """Pop the last compound id in this set"""
        return self._indices.pop()

    def shuffle(self) -> None:
        """Randomises the order of compounds in this set"""
        from random import shuffle

        shuffle(self._indices)

    def get_df(
        self,
        smiles: bool = True,
        inchikey: bool = False,
        alias: bool = True,
        mol: bool = False,
        metadata: bool = False,
        expand_metadata: bool = True,
        poses: bool = False,
        num_reactant: bool = False,
        num_reactions: bool = False,
        num_poses: bool = False,
        tags: bool = False,
        bases: bool = False,
        elabs: bool = False,
        routes: bool = False,
        debug: bool = False,
        # count_by_target: bool = False,
        **kwargs,
    ) -> "DataFrame":
        """Get a DataFrame representation of this set

        :param smiles: include SMILES column (Default value = True)
        :param inchikey: include InChIKey column (Default value = False)
        :param alias: include alias column (Default value = True)
        :param mol: include ``rdkit.Chem.Mol`` in output (Default value = False)
        :param metadata: include metadata in output (Default value = False)
        :param expand_metadata: create separate column for each metadata key (Default value = True)
        :param poses: include poses in output (Default value = False)
        :param num_reactant: include num_poses column
        :param num_reactant: include num_reactant column (number of reactions where compound is a reactant)
        :param num_reactions: include num_reactions column (number of reactions where compound is a product)
        :param tags: include tags column
        :param bases: include bases column
        :param elabs: include elabs column

        # :param count_by_target: count poses by target (Default value = False)

        """

        from json import loads
        from rdkit.Chem import Mol
        from pandas import DataFrame

        data = []

        query = ["compound_id"]

        if smiles:
            query.append("compound_smiles")

        if inchikey:
            query.append("compound_inchikey")

        if alias:
            query.append("compound_alias")

        if mol:
            query.append("mol_to_binary_mol(compound_mol)")

        if metadata:
            query.append("compound_metadata")

        query = ", ".join(query)

        sql = f"""
        SELECT {query}
        FROM compound
        WHERE compound_id IN {self.str_ids}
        """

        if debug:
            mrich.debug("querying...")
        records = self.db.execute(sql).fetchall()

        if debug:
            generator = mrich.track(records)
        else:
            generator = records

        for row in generator:

            row = list(row)

            d = dict(id=row.pop(0))

            if smiles:
                d["smiles"] = row.pop(0)

            if inchikey:
                d["inchikey"] = row.pop(0)

            if alias:
                d["alias"] = row.pop(0)

            if mol:
                d["mol"] = Mol(row.pop(0))

            if metadata and (meta_str := row.pop(0)):

                meta_dict = loads(meta_str)

                if expand_metadata:
                    for k, v in meta_dict.items():
                        d[k] = v

                else:
                    d["metadata"] = meta_dict

            data.append(d)

        df = DataFrame(data)

        if poses or num_poses:
            if debug:
                mrich.debug("adding pose column")

            lookup = self.db.get_compound_id_pose_ids_dict(self)
            if poses:
                df["poses"] = df["id"].apply(lambda x: lookup.get(x, {}))
            if num_poses:
                df["num_poses"] = df["id"].apply(lambda x: len(lookup.get(x, {})))

        if num_reactant or num_reactions:
            if debug:
                mrich.debug("adding reaction columns")
            tuples = self.db.get_reactant_product_tuples(self.ids, deduplicated=False)

            if num_reactant:
                lookup = {}
                for r, p in tuples:
                    lookup.setdefault(r, 0)
                    lookup[r] += 1
                df["num_reactant"] = df["id"].apply(lambda x: lookup.get(x, 0))

            if num_reactions:
                lookup = {}
                for r, p in tuples:
                    lookup.setdefault(p, 0)
                    lookup[p] += 1
                df["num_reactions"] = df["id"].apply(lambda x: lookup.get(x, 0))

        if bases or elabs:
            if debug:
                mrich.debug("adding scaffold columns")
            tuples = self.db.get_scaffold_tuples(self.ids)

            if bases:
                lookup = {}
                for b, e in tuples:
                    lookup.setdefault(e, set())
                    lookup[e].add(b)
                df["bases"] = df["id"].apply(lambda x: lookup.get(x, set()))

            if elabs:
                lookup = {}
                for b, e in tuples:
                    lookup.setdefault(b, set())
                    lookup[b].add(e)
                df["elabs"] = df["id"].apply(lambda x: lookup.get(x, set()))

        if tags:
            if debug:
                mrich.debug("adding tag column")
            lookup = self.db.get_compound_tag_dict()
            df["tags"] = df["id"].apply(lambda x: lookup.get(x, {}))

        if routes:
            if debug:
                mrich.debug("adding route column")
            lookup = self.db.get_product_id_routes_dict()
            df["routes"] = df["id"].apply(lambda x: lookup.get(x, {}))

        df = df.set_index("id")

        return df

    def get_quoted(
        self,
        *,
        supplier: str = "any",
    ) -> "CompoundSet":
        """Get all member compounds that have a quote from given supplier

        :param supplier: supplier name (Default value = 'any')

        """

        if supplier == "any":
            key = f"quote_compound IN {self.str_ids}"
        else:
            key = f'quote_compound IN {self.str_ids} AND quote_supplier = "{supplier}"'

        ids = self.db.select_where(
            table="quote",
            query="DISTINCT quote_compound",
            key=key,
            multiple=True,
        )

        ids = [i for i, in ids]
        return CompoundSet(self.db, ids)

    def get_unquoted(
        self,
        *,
        supplier: str = "any",
    ) -> "CompoundSet":
        """Get all member compounds that do not have a quote from given supplier

        :param supplier: supplier name (Default value = 'any')

        """

        quoted = self.get_quoted(supplier=supplier)
        return self - quoted

    def get_dict(self) -> dict:
        """Get a dictionary object with all serialisable data needed to reconstruct this set"""
        return dict(db=str(self.db.path.resolve()), indices=self.indices)

    def write_smiles_csv(self, file: str, tags: bool = True) -> None:
        """Write a CSV of the smiles contained in this set to a file

        :param file: path of the CSV file

        """
        from pandas import DataFrame

        if tags:
            records = self.db.select_where(
                table="tag",
                query="tag_compound, tag_name",
                key=f"tag_compound IN {self.str_ids}",
                multiple=True,
                none="quiet",
            )
            TAGS = {}
            if records:
                for compound_id, tag_name in records:
                    if compound_id not in TAGS:
                        TAGS[compound_id] = set()
                    TAGS[compound_id].add(tag_name)

        records = self.db.select_where(
            table=self.table,
            query="compound_id, compound_smiles",
            key=f"compound_id IN {self.str_ids}",
            multiple=True,
        )

        if tags:
            data = [
                dict(id=id, smiles=smiles, tags=TAGS.get(id, []))
                for id, smiles in records
            ]
        else:
            data = [dict(id=id, smiles=smiles) for id, smiles in records]

        df = DataFrame(data)
        mrich.writing(file)
        df.to_csv(file, index=False)

    def write_postera_csv(
        self,
        file,
        *,
        supplier: str = "Enamine",
        prefix: str = "fragment",
    ) -> None:
        """Write a CSV formatted for upload to Postera's Manifold

        :param file: path of the CSV file
        :param supplier: supplier to use for quotes, (Default value = 'Enamine')
        :param prefix: prefix to metadata columns, (Default value = 'fragment')

        """

        from datetime import date as dt
        from pandas import DataFrame

        if prefix:
            prefix = f"{prefix}_"

        data = []

        for c in mrich.track(self, prefix="Creating DataFrame"):

            # get props
            smiles = c.smiles
            tags = c.tags
            metadata = c.metadata
            poses = c.poses
            base = c.base

            # method
            assert len(tags) == 1, c
            method = tags[0]

            # date
            date = dt.today()

            # author
            assert "author" in metadata, c
            author = metadata["author"]

            match len(poses):
                case 1:
                    pose = poses[0]
                case 0:
                    mrich.warning(f"{c} has no poses")
                    assert base
                    pose = base.poses[0]
                case _:
                    mrich.warning(f"{c} has multiple poses")
                    pose = poses[0]

            # extract inspirations
            inspirations = pose.inspirations
            inspiration_names = ",".join(inspirations.names)
            inspiration_smiles = ".".join(inspirations.smiles)

            # quote info
            quotes = c.get_quotes(supplier=supplier)
            assert len(quotes) == 1, c
            quote = quotes[0]
            catalog_id = quote.entry
            catalog_price = quote.price
            catalog_lead_time = quote.lead_time

            # hippo string
            hippo_str = f"compound={c.id}, pose={pose.id}"

            # create row
            data.append(
                {
                    "SMILES": smiles,
                    f"{prefix}HIPPO_IDs": hippo_str,
                    f"{prefix}method": method,
                    f"{prefix}export_date": date,
                    f"{prefix}author": author,
                    f"{prefix}inspiration_names": inspiration_names,
                    f"{prefix}inspiration_SMILES": inspiration_smiles,
                    f"{prefix}supplier": supplier,
                    f"{prefix}supplier_catalogue": quote.catalogue,
                    f"{prefix}supplier_ID": catalog_id,
                    f"{prefix}supplier_price": catalog_price,
                    f"{prefix}supplier_lead_time": catalog_lead_time,
                }
            )

        df = DataFrame(data)

        mrich.writing(file)
        df.to_csv(file, index=False)

        return df

    def write_CAR_csv(
        self,
        file: "str | Path",
        amount: float = 1,  # in mg
        return_df: bool = False,
        # pick_cheapest: bool = False,
        quoted_only: bool = False,
        get_ingredient_quotes: bool = True,
        **kwargs,
    ) -> "DataFrame | None":
        """List of reactions for CAR

        Columns:

        * target-name
        * no-steps
        * concentration = None
        * amount-required
        * batch-tag

        per reaction

        * reactant-1-1
        * reactant-2-1
        * reaction-product-smiles-1
        * reaction-name-1
        * reaction-recipe-1
        * reaction-groupby-column-1

        :param file: output file
        :param amount: amount of each product in `mg`
        :param quoted_only: only choose reactants that have quotes
        :param supplier: only choose reactants that have quotes from this supplier
        :param kwargs: passed to :meth:`.Recipe.from_reaction`
        :param return_df: return a `DataFrame` (Default value = False)

        """

        from pathlib import Path
        from pandas import DataFrame
        from .recipe import Recipe

        file = str(Path(file).resolve())

        rows = []

        for r_id in mrich.track(self.reaction_ids, prefix="Solving compound recipes"):

            reaction = self.db.get_reaction(id=r_id)

            recipes = Recipe.from_reaction(
                reaction,
                amount=amount,
                pick_cheapest=False,
                quoted_only=quoted_only,
                get_ingredient_quotes=get_ingredient_quotes,
                **kwargs,
            )

            for sub_recipe in recipes:

                product = sub_recipe.product

                row = {
                    "target-names": str(product.compound),
                    "no-steps": 0,
                    "concentration-required-mM": None,
                    "amount-required-uL": None,
                    "batch-tag": None,
                }

                for i, reaction in enumerate(sub_recipe.reactions):

                    i = i + 1

                    row["no-steps"] += 1

                    match len(reaction.reactants):
                        case 1:
                            row[f"reactant-1-{i}"] = reaction.reactants[0].smiles
                            row[f"reactant-2-{i}"] = None
                        case 2:
                            row[f"reactant-1-{i}"] = reaction.reactants[0].smiles
                            row[f"reactant-2-{i}"] = reaction.reactants[1].smiles
                        case _:
                            raise NotImplementedError(
                                f"Unsupported number of reactants for {reaction=}: {len(reaction.reactants)}"
                            )

                    row[f"reaction-product-smiles-{i}"] = reaction.product.smiles
                    row[f"reaction-name-{i}"] = reaction.type
                    row[f"reaction-recipe-{i}"] = None
                    row[f"reaction-groupby-column-{i}"] = None
                    # row[f'reaction-id-{i}'] = int(reaction.id)

                rows.append(row)

        df = DataFrame(rows)

        df = df.convert_dtypes()

        for n_steps in set(df["no-steps"]):
            subset = df[df["no-steps"] == n_steps]
            this_file = file.replace(".csv", f"_{n_steps}steps.csv")
            mrich.writing(this_file)
            subset.to_csv(this_file, index=False)

        mrich.writing(file)
        df.to_csv(file, index=False)

        if return_df:
            return df

    def add_tag(
        self,
        tag: str,
    ) -> None:
        """Add this tag to every member of the set"""

        assert isinstance(tag, str)

        for i in self.indices:
            self.db.insert_tag(name=tag, compound=i, commit=False)

        mrich.print(f'Tagged {self} w/ "{tag}"')

        self.db.commit()

    def plot_tsnee(self, **kwargs) -> "go.Figure":
        """Plot a tanimoto similarity plot of these compounds"""
        from .plotting import plot_compound_tsnee

        return plot_compound_tsnee(self, **kwargs)

    ### DUNDERS

    def __len__(self) -> int:
        """The number of compounds in this set"""
        return len(self.indices)

    def __iter__(self):
        """Iterate through compounds in this set"""
        return iter(self.db.get_compound(id=i) for i in self.indices)

    def __getitem__(
        self,
        key: int | slice,
    ) -> "Compound | CompoundSet":
        """Get compounds or subsets thereof from this set

        :param key: integer index or slice of indices

        """
        match key:
            case int():
                index = self.indices[key]
                return self.db.get_compound(id=index)

            case slice():
                indices = self.indices[key]
                return CompoundSet(self.db, indices)

            case _:
                raise NotImplementedError

    def __sub__(
        self,
        other: "CompoundSet | IngredientSet",
    ) -> "CompoundSet":
        """Subtract a :class:`.Compound` object or ID to this set, or subtract multiple at once when ``other`` is a :class:`.CompoundSet` or :class:`.IngredientSet`"""

        match other:

            case CompoundSet():
                ids = set(self.ids) - set(other.ids)
                return CompoundSet(self.db, ids)

            case IngredientSet():
                mrich.warning(
                    "Subtracting IngredientSet from CompoundSet. Ignoring quote/amount data"
                )
                ids = set(self.ids) - set([int(i) for i in other.compound_ids])
                return CompoundSet(self.db, ids)

            case _:
                raise NotImplementedError

    def __add__(
        self,
        other: "Compound | CompoundSet | IngredientSet | int",
    ) -> "CompoundSet":
        """Add a :class:`.Compound` object or ID to this set, or add multiple at once when ``other`` is a :class:`.CompoundSet` or :class:`.IngredientSet`"""

        match other:

            case Compound():
                return self.add(other)

            case int():
                return self.add(other)

            case CompoundSet():
                ids = set(self.ids) | set(other.ids)
                return CompoundSet(self.db, ids)

            case IngredientSet():
                ids = set(self.ids) | set(other.compound_ids)
                return CompoundSet(self.db, ids)

            case _:
                raise NotImplementedError

    def __xor__(self, other: "CompoundSet"):
        """Exclusive OR set operation, returns all compounds in either set but not both"""

        match other:

            case CompoundSet():
                ids = set(self.ids) ^ set(other.ids)
                return CompoundSet(self.db, ids)

            case _:
                raise NotImplementedError

    def __str__(self) -> str:
        """Unformatted string representation"""

        if self.name:
            s = f"{self.name}: "
        else:
            s = ""

        s += "{" f"C × {len(self)}" "}"

        return s

    def __repr__(self) -> str:
        """ANSI ormatted string representation"""
        return f"{mcol.bold}{mcol.underline}{self}{mcol.unbold}{mcol.ununderline}"

    def __rich__(self) -> str:
        """Representation for mrich"""
        return f"[bold underline]{self}"

    def __contains__(self, other: Compound | Ingredient | int):
        """Check if compound or ingredient is a member of this set"""
        match other:
            case Compound():
                id = other.id
            case Ingredient():
                id = other.compound_id
            case int():
                id = other

        return id in set(self.ids)


class IngredientSet:
    """An :class:`.Ingredient` is a :class:`.Compound` with a fixed quanitity and an attached quote, the :class:`.IngredientSet` is a object representing multiple ingredients.

    .. attention::

            :class:`.IngredientSet` objects should not be created directly. Instead they are returned by several methods when working with :doc:`quoting` and :doc:`rgen`.

    Selecting ingredients in the set
    ================================

    The :class:`.IngredientSet` can be indexed like a Python list:

    ::

            ingredient = ingredient_set[0] # first ingredient

    To get the ingredient for a specific :class:`.Compound` ID:

    ::

            ingredient = ingredient_set(compound_id=13)

    """

    _columns = [
        "compound_id",
        "amount",
        "quote_id",
        "supplier",
        "max_lead_time",
        "quoted_amount",
    ]

    def __init__(
        self,
        db: "Database",
        ingredients: "None | list[Ingredient]" = None,
        supplier: str | list | None = None,
        debug: bool = False,
    ) -> None:

        from pandas import DataFrame

        ingredients = ingredients or []

        self._db = db

        self._data = DataFrame(columns=self._columns, dtype=object)

        if debug:
            mrich.debug(self._data)

        self._supplier = supplier

        for ingredient in ingredients:
            self.add(ingredient)

        for col in self._columns:
            assert col in self._data.columns, f"{col} not in df.columns"

        if debug:
            mrich.debug(self._data)

    @classmethod
    def from_ingredient_df(
        cls,
        db: "Database",
        df: "DataFrame",
        supplier: str | list | None = None,
    ) -> "IngredientSet":
        """Create an :class:`.IngredientSet` from a DataFrame

        :param db: HIPPO Database
        :param df: DataFrame of Ingredients
        :param supplier: supplier to use for all quoting, (Default value = None)

        """
        # from numpy import nan
        self = cls.__new__(cls)

        for col in cls._columns:
            if col not in df.columns:
                raise Exception(f"{col} not in df.columns")
                df[col] = None

        self._db = db
        self._data = df.copy()
        self._supplier = supplier

        return self

    @classmethod
    def from_json(
        cls,
        db: "Database",
        path: None | str,
        supplier: str | list | None = None,
        data: None | dict = None,
    ) -> "IngredientSet":
        """Create an :class:`.IngredientSet` from JSON data or a JSON file

        :param db: HIPPO Database
        :param path: path to JSON data (can be ``None`` if ``data`` provided)
        :param supplier: supplier to use for all quoting, (Default value = ``None``)
        :param data: optional JSON data to parse, (Default value = ``None``)

        """

        if not data:
            import json

            data = json.load(open(path, "rt"))

        from pandas import DataFrame

        df = DataFrame(columns=cls._columns, dtype=object)

        for col in cls._columns:
            df[col] = data[col]

        return cls.from_ingredient_df(db=db, df=df, supplier=supplier)

    @classmethod
    def from_ingredient_dicts(
        cls,
        db: "Database",
        dicts: list[dict],
        supplier: str | list | None = None,
    ) -> "IngredientSet":
        """Create an :class:`.IngredientSet` from :class:`.Ingredient` dictionaries

        :param db: HIPPO Database
        :param dicts: List of individual ingredient dictionaries
        :param supplier: supplier to use for all quoting, (Default value = ``None``)

        """
        from pandas import DataFrame

        df = DataFrame(dicts, dtype=object)
        return cls.from_ingredient_df(db=db, df=df, supplier=supplier)

    @classmethod
    def from_compounds(
        cls,
        *,
        compounds: "CompoundSet | None" = None,
        ids: list[int] | None = None,
        db: "Database | None" = None,
        amount: float | list[float] = 1,
        supplier: str | list | None = None,
    ) -> "IngredientSet":
        """Create an :class:`.IngredientSet` from a :class:`.CompoundSet` or IDs

        :param compounds: :class:`.CompoundSet` to use, if ``None`` must provide ``ids`` and ``db`` (Default value = None)
        :param ids: Compound IDs (Default value = None)
        :param db: HIPPO Database (Default value = None)
        :param amount: Amount(s) in ``mg`` (Default value = 1)
        :param supplier: supplier to use for all quoting, (Default value = ``None``)

        """

        from pandas import DataFrame

        if not ids:
            ids = compounds.ids

        if not db:
            db = compounds.db

        df = DataFrame(
            dict(
                compound_id=ids,
                amount=amount,
                quote_id=None,
                supplier=supplier,
                max_lead_time=None,
                quoted_amount=None,
            ),
            dtype=object,
        )

        return cls.from_ingredient_df(db, df)

    ### PROPERTIES

    @property
    def df(self) -> "DataFrame":
        """Access the raw DataFrame"""
        return self._data

    @property
    def db(self) -> "Database":
        """Linked HIPPO Database"""
        return self._db

    @property
    def price_df(self) -> "DataFrame":
        """DataFrame including prices"""
        df = self.df.copy()
        df["price"] = [i.price for i in self]
        return df

    @property
    def price(self) -> "Price":
        """Total price of these ingredients"""
        return self.get_price()

    @property
    def supplier(self) -> str | list[str]:
        """Supplier(s)"""
        return self._supplier

    @supplier.setter
    def supplier(self, s):

        if isinstance(s, list) or isinstance(s, tuple):
            for x in s:
                assert isinstance(x, str)
        else:
            assert isinstance(s, str)

        self._supplier = s
        self.df["supplier"] = [s] * len(self)

    @property
    def smiles(self) -> list[str]:
        """SMILES for all ingredients"""
        compound_ids = list(self.df["compound_id"])
        result = self.db.select_where(
            query="compound_smiles",
            table="compound",
            key=f"compound_id in {tuple(compound_ids)}",
            multiple=True,
        )
        return [q for q, in result]

    @property
    def inchikeys(self) -> list[str]:
        """InChI-keys for all ingredients"""
        compound_ids = list(self.df["compound_id"])
        result = self.db.select_where(
            query="compound_inchikey",
            table="compound",
            key=f"compound_id in {tuple(compound_ids)}",
            multiple=True,
        )
        return [q for q, in result]

    @property
    def compound_ids(self) -> list[int]:
        """Compound IDs for all ingredients"""
        return list(self.df["compound_id"].values)

    @property
    def ids(self) -> list[int]:
        """Compound IDs for all ingredients"""
        return self.compound_ids

    @property
    def id_amount_pairs(self) -> list[tuple]:
        """Get a list of compound ID and amount pairs"""
        return [
            (id, amount) for id, amount in self.df[["compound_id", "amount"]].values
        ]

    @property
    def str_compound_ids(self) -> str:
        """Return an SQL formatted tuple string of the :class:`.Compound` IDs"""
        return str(tuple(self.df["compound_id"].values)).replace(",)", ")")

    @property
    def compounds(self) -> "CompoundSet":
        """:class:`.CompoundSet` of all compounds in this set"""
        return CompoundSet(self.db, self.compound_ids)

    ### METHODS

    def get_price(
        self, supplier: str | list[str] = None, none: str = "error", debug: bool = False
    ) -> "Price":
        """Calculate the price with a given supplier

        :param supplier: supplier to use for all quoting, (Default value = ``None``)

        """

        from .price import Price

        pairs = {i: q for i, q in enumerate(self.df["quote_id"])}

        quote_ids = [q for q in pairs.values() if q is not None and not isnan(q)]

        if debug:
            mrich.debug("quote_ids", quote_ids)

        if quote_ids:

            quote_id_str = str(tuple(quote_ids)).replace(",)", ")")

            if supplier:
                result = self.db.select_where(
                    query="quote_price, quote_currency",
                    table="quote",
                    key=f'quote_id in {quote_id_str} AND quote_supplier = "{supplier}"',
                    multiple=True,
                    none=none,
                )
            else:
                result = self.db.select_where(
                    query="quote_price, quote_currency",
                    table="quote",
                    key=f"quote_id in {quote_id_str}",
                    multiple=True,
                    none=none,
                )

            prices = [Price(a, b) for a, b in result]
            quoted = sum(prices, Price.null())

        else:

            quoted = Price.null()

        if debug:
            mrich.debug("quoted", quoted)

        unquoted = [i for i, q in pairs.items() if q is None or isnan(q)]

        unquoted_price = Price.null()

        for i in unquoted:

            ingredient = self[i]

            if debug:
                mrich.debug("unquoted", i, ingredient)

            p = ingredient.price

            unquoted_price += p

            if debug:
                mrich.debug(unquoted_price)

            quote = ingredient.quote

            if not quote:
                mrich.warning("NULL Quote:", ingredient)
                continue

            self.df.loc[i, "quote_id"] = quote.id

            assert quote.amount

            self.df.loc[i, "quoted_amount"] = quote.amount

        if debug:
            mrich.debug("quoted", quoted)
            mrich.debug("unquoted_price", unquoted_price)
            mrich.error("end of IngredientSet.get_price()")

        return quoted + unquoted_price

    def interactive(self, **kwargs) -> None:
        """Wrapper for :meth:`.CompoundSet.interactive`"""
        self.compounds.interactive(**kwargs)

    def add(
        self,
        ingredient: "Ingredient | None" = None,
        *,
        compound_id: int | None = None,
        amount: float | None = None,
        quote_id: int | None = None,
        supplier: str | list[str] | None = None,
        max_lead_time: float | None = None,
        quoted_amount: float | None = None,
        debug: bool = False,
    ) -> None:
        """Add an :class:`.Ingredient` to this set

        :param ingredient: :class:`.Ingredient` to be added, if ``None`` must specify other parameters, (Default value = None)
        :param compound_id: :class:`.Compound` ID (Default value = None)
        :param amount: amount in ``mg`` (Default value = None)
        :param quote_id: :class:`.Quote` ID (Default value = None)
        :param supplier: supplier name string or list (Default value = None)
        :param max_lead_time: maximum lead-time for quoting (in days) (Default value = None)
        :param quoted_amount: amount of associated :class:`.Quote` (Default value = None)
        :param debug: increase verbosity for debugging (Default value = False)

        """

        from pandas import DataFrame, concat

        if ingredient:
            assert ingredient._table == "ingredient"
            compound_id = ingredient.compound_id
            amount = ingredient.amount

            if (q := ingredient.quote) and not ingredient.quote_id:
                mrich.warning(f"Losing quote! {ingredient.quote=}")

            supplier = ingredient.supplier
            max_lead_time = ingredient.max_lead_time

            if q is None:
                quote_id = None
                quoted_amount = None
            else:
                quote_id = q.id
                quoted_amount = q.amount

        else:
            assert compound_id
            assert amount

        if quote_id:
            # if not quoted_amount:
            #     mrich.warning(f'Requoting C{compound_id}...')

            assert quoted_amount

        supplier = self.supplier

        if self._data.empty:
            addition = DataFrame(
                [
                    dict(
                        compound_id=compound_id,
                        amount=amount,
                        quote_id=quote_id,
                        supplier=supplier,
                        max_lead_time=max_lead_time,
                        quoted_amount=quoted_amount,
                    )
                ],
                dtype=object,
            )
            self._data = addition

        else:

            if compound_id in self._data["compound_id"].values:
                index = self._data.index[
                    self._data["compound_id"] == compound_id
                ].tolist()[0]
                self._data.loc[index, "amount"] += amount

                # discard if the quote is no longer valid
                if (a := self.df.loc[index, "quoted_amount"]) and a < self.df.loc[
                    index, "amount"
                ]:
                    self._data.loc[index, "quote_id"] = None
                    self._data.loc[index, "quoted_amount"] = None

                if debug and supplier:
                    mrich.debug("Adding to existing ingredient")
                    mrich.debug(f'{self._data.loc[index, "supplier"]=}')
                    mrich.debug(f"{supplier=}")

            else:
                # from numpy import nan
                addition = DataFrame(
                    [
                        dict(
                            compound_id=compound_id,
                            amount=amount,
                            quote_id=quote_id,
                            supplier=supplier,
                            max_lead_time=max_lead_time,
                            quoted_amount=quoted_amount,
                        )
                    ],
                    dtype=object,
                )

                self._data = concat(
                    [self._data, addition], ignore_index=True, join="inner"
                )

                if debug:
                    mrich.out(addition)

    def _get_ingredient(
        self,
        series,
    ) -> "Ingredient":
        """Get ingredient from one of the DataFrame rows"""

        q_id = series["quote_id"]

        if isinstance(q_id, float) and isnan(q_id):
            q_id = None

        return Ingredient(
            db=self._db,
            compound=series["compound_id"],
            amount=series["amount"],
            quote=q_id,
            supplier=series["supplier"],
            max_lead_time=series["max_lead_time"],
        )

    def copy(self) -> "IngredientSet":
        """Return a copy of this :class:`.IngredientSet`"""
        return IngredientSet.from_ingredient_df(
            self.db, self.df, supplier=self.supplier
        )

    def draw(self) -> None:
        """Wrapper for :meth:`.CompoundSet.draw`"""
        self.compounds.draw()

    def set_amounts(
        self,
        amount: float | list[float],
    ) -> None:
        """Set the amount(s) for all ingredients in this set, and update quotes

        :param amount: amount in ``mg``

        """

        self.df["amount"] = amount

        # if amounts are modified the quotes should be cleared
        self.df["quote_id"] = None

        assert all(self.df["supplier"].isna()) and all(self.df["max_lead_time"].isna())

        # update quotes
        pairs = self.db.execute(
            f"""
            WITH matching_quotes AS (
                SELECT quote_id, quote_compound, MIN(quote_price) FROM quote
                WHERE quote_compound IN {self.str_compound_ids}
                AND quote_amount >= {amount}
                GROUP BY quote_compound
            )
            SELECT compound_id, quote_id FROM compound
            LEFT JOIN matching_quotes ON quote_compound = compound_id
            WHERE compound_id IN {self.str_compound_ids}
        """
        ).fetchall()

        for compound_id, quote_id in pairs:
            match = self.df.index[self.df["compound_id"] == compound_id][0]
            self.df.loc[match, "quote_id"] = quote_id

    def get_dict(self, data_orient: str = "list") -> dict:
        """Get serialisable dictionary

        :param data_orient: passed to ``pandas.DataFrame.to_dict`` (Default value = 'list')

        """
        return dict(
            db=str(self.db),
            supplier=self.supplier,
            data=self.df.to_dict(orient=data_orient),
        )

    def pop(self) -> Ingredient:
        """Pop the last compound in this set"""
        item = self[self.df.index[-1]]
        self.df.drop(self.df.index[-1], inplace=True)
        return item

    def shuffle(self) -> None:
        """Randomises the order of compounds in this set"""
        self._data = self.df.sample(frac=1).reset_index(drop=True)

    ### DUNDERS

    def __len__(self):
        """The number of ingredients in this set"""
        return len(self._data)

    def __str__(self) -> str:
        """Unformatted string representation"""
        return "{" f"Ingredient × {len(self)}" "}"

    def __repr__(self) -> str:
        """ANSI ormatted string representation"""
        return f"{mcol.bold}{mcol.underline}{self}{mcol.unbold}{mcol.ununderline}"

    def __rich__(self) -> str:
        """Representation for mrich"""
        return f"[bold underline]{self}"

    def __add__(self, other):
        """Add another  :class:`.IngredientSet` this set"""

        for i, row in other._data.iterrows():
            self.add(
                compound_id=row.compound_id,
                amount=row.amount,
                quote_id=row.quote_id,
                supplier=row.supplier,
                max_lead_time=row.max_lead_time,
                quoted_amount=row.quoted_amount,
            )

        return self

    def __getitem__(self, key: int) -> "Ingredient":
        """Get a member by it's index"""
        match key:
            case int():
                series = self.df.loc[key]
                return self._get_ingredient(series)

            case _:
                raise NotImplementedError

    def __iter__(self):
        """Iterate through the ingredients"""
        return iter(self._get_ingredient(s) for i, s in self.df.iterrows())

    def __call__(
        self,
        *,
        compound_id: int | None = None,
        tag: str | None = None,
    ) -> "IngredientSet | Ingredient | CompoundSet":
        """Get members based on a compound_id or tag"""

        if compound_id:

            # get the ingredient with the matching compound ID
            matches = self.df[self.df["compound_id"] == compound_id]

            if len(matches) == 0:
                return None

            elif len(matches) != 1:

                mrich.warning(f"Multiple ingredients in set with {compound_id=}")
                # print(matches)

                return IngredientSet(
                    self.db, [self._get_ingredient(s) for i, s in matches.iterrows()]
                )

            return self._get_ingredient(matches.iloc[0])

        # elif tag:
        #     return self.compounds(tag=tag)

        else:
            raise NotImplementedError

    def __getattr__(self, key: str):
        """For missing attributes try getting from associated :class:`.CompoundSet`"""
        return getattr(self.compounds, key)

    def __contains__(self, other: Compound | Ingredient | int):
        """Check if compound or ingredient is a member of this set"""
        match other:
            case Compound():
                id = other.id
            case Ingredient():
                id = other.compound_id
            case int():
                id = other

        return id in set(self.compound_ids)



import mcol

from rdkit import Chem

from .pose import Pose

# from .pset import PoseSet
from .tags import TagSet

# from .rset import ReactionSet
from .target import Target
from .quote import Quote

import mrich


class Compound:
    """A :class:`.Compound` represents a ligand/small molecule with stereochemistry removed and no atomic coordinates. I.e. it represents the chemical structure. It's name is always an InChiKey. If a compound is an elaboration it can have a :meth:`.Compound.bases` property which is another :class:`.Compound`. :class:`.Compound` objects are target-agnostic and can be linked to any number of catalogue entries (:class:`.Quote`) or synthetic pathways (:class:`.Reaction`).

    .. attention::

            :class:`.Compound` objects should not be created directly. Instead use :meth:`.HIPPO.register_compound` or :meth:`.HIPPO.compounds`. See :doc:`getting_started` and :doc:`insert_elaborations`.

    """

    _table = "compound"

    def __init__(
        self,
        animal: "HIPPO",
        db: "Database",
        id: int,
        inchikey: str,
        alias: str,
        smiles: str,
        mol: Chem.Mol | bytes | None = None,
        metadata: dict | None = None,
    ):

        # from compound table
        self._id = id
        self._inchikey = inchikey
        self._alias = alias
        self._smiles = smiles
        self._animal = animal
        self._bases = None
        self._elabs = None
        self._alias = alias
        self._tags = None
        self._metadata = metadata

        # computed properties
        self._num_heavy_atoms = None
        self._num_rings = None
        self._formula = None
        self._molecular_weight = None
        self._total_changes = db.total_changes

        if isinstance(mol, bytes):
            mol = Chem.Mol(mol)

        self._mol = mol

        self._db = db

    ### FACTORIES

    ### PROPERTIES

    @property
    def id(self) -> int:
        """Returns the compound's database ID"""
        return self._id

    @property
    def inchikey(self) -> str:
        """Returns the compound's InChiKey"""
        return self._inchikey

    @property
    def name(self) -> str:
        """Returns the compound's InChiKey"""
        if self.alias:
            return self.alias
        return self.inchikey

    @property
    def smiles(self) -> str:
        """Returns the compound's (flattened) smiles"""
        return self._smiles

    @property
    def alias(self) -> str:
        """Returns the compound's alias"""
        return self._alias

    @alias.setter
    def alias(self, alias: str) -> None:
        """Set the compound's alias"""
        self.set_alias(alias)

    @property
    def mol(self) -> Chem.Mol:
        """Returns the compound's RDKit Molecule"""
        if self._mol is None:
            (mol,) = self.db.select_where(
                query="mol_to_binary_mol(compound_mol)",
                table="compound",
                key="id",
                value=self.id,
                multiple=False,
            )
            self._mol = Chem.Mol(mol)
        return self._mol

    @property
    def num_heavy_atoms(self) -> int:
        """Get the number of heavy atoms"""
        if self._num_heavy_atoms is None:
            self._num_heavy_atoms = self.db.get_compound_computed_property(
                "num_heavy_atoms", self.id
            )
        return self._num_heavy_atoms

    @property
    def molecular_weight(self) -> float:
        """Get the molecular weight"""
        if self._molecular_weight is None:
            self._molecular_weight = self.db.get_compound_computed_property(
                "molecular_weight", self.id
            )
        return self._molecular_weight

    @property
    def num_rings(self) -> int:
        """Get the number of rings"""
        if self._num_rings is None:
            self._num_rings = self.db.get_compound_computed_property(
                "num_rings", self.id
            )
        return self._num_rings

    @property
    def formula(self) -> str:
        """Get the chemical formula"""
        if self._formula is None:
            self._formula = self.db.get_compound_computed_property("formula", self.id)
        return self._formula

    @property
    def atomtype_dict(self) -> dict[str, int]:
        """Get a dictionary with atomtypes as keys and corresponding quantities/counts as values."""
        from molparse.atomtypes import formula_to_atomtype_dict

        return formula_to_atomtype_dict(self.formula)

    @property
    def num_atoms_added(self) -> int | list[int] | None:
        """Calculate the number of atoms added relative to the base compound"""
        match self.num_bases:
            case 0:
                mrich.error(f"{self} has no base")
                return None
            case 1:
                b_id = self.bases.ids[0]
                n_e = self.num_heavy_atoms
                n_b = self.db.get_compound_computed_property("num_heavy_atoms", b_id)
                return n_e - n_b
            case _:
                mrich.warning(f"{self} has multiple bases")
                n_e = self.num_heavy_atoms
                return [
                    n_e
                    - self.db.get_compound_computed_property("num_heavy_atoms", b_id)
                    for b_id in self.bases.ids
                ]

    @property
    def metadata(self) -> "MetaData":
        """Returns the compound's metadata dict"""
        if self._metadata is None:
            self._metadata = self.db.get_metadata(table="compound", id=self.id)
        return self._metadata

    @property
    def db(self) -> "Database":
        """Returns a pointer to the parent database"""
        return self._db

    @property
    def tags(self) -> TagSet:
        """Returns the compound's tags"""
        if not self._tags:
            self._tags = self.get_tags()
        return self._tags

    @property
    def poses(self) -> "PoseSet":
        """Returns the compound's poses"""
        return self.get_poses()

    @property
    def best_placed_pose(self) -> Pose:
        """Returns the compound's pose with the lowest distance score"""
        return self.poses.best_placed_pose

    @property
    def num_poses(self) -> int:
        """Returns the number of associated poses"""
        return self.db.count_where(table="pose", key="compound", value=self.id)

    @property
    def num_reactions(self) -> int:
        """Returns the number of associated reactions (product)"""
        return self.db.count_where(table="reaction", key="product", value=self.id)

    @property
    def num_reactant(self) -> int:
        """Returns the number of associated reactions (reactant)"""
        return self.db.count_where(table="reactant", key="compound", value=self.id)

    @property
    def bases(self) -> "CompoundSet | None":
        """Returns the base compound for this elaboration"""
        if self._bases is None or self._db_changed:
            ids = self.get_base_ids()
            if not ids:
                self._bases = None
            else:
                from .cset import CompoundSet

                self._bases = CompoundSet(
                    self.db, ids, name=f"base scaffolds of {self}"
                )
                self._total_changes = self.db.total_changes
        return self._bases

    @property
    def num_bases(self) -> int:
        """Get the number of base compounds for this elaboration"""
        return len(self.bases)

    @property
    def elabs(self):
        """Returns the base compound for this elaboration"""
        if self._elabs is None or self._db_changed:
            ids = self.get_superstructure_ids()
            if not ids:
                self._elabs = None
            else:
                from .cset import CompoundSet

                self._elabs = CompoundSet(self.db, ids, name=f"elaborations of {self}")
                self._total_changes = self.db.total_changes
        return self._elabs

    @property
    def reactions(self) -> "ReactionSet":
        """Returns the reactions resulting in this compound"""
        return self.get_reactions(none=False)

    @property
    def reaction(self) -> "Reaction":
        """Returns the reaction resulting in this compound (will return first if multiple, with a warning)"""
        reactions = self.reactions
        match len(reactions):
            case 0:
                mrich.warning(f"{self} has no reactions")
                return None
            case 1:
                mrich.warning(f"{self} has multiple reactions, returning first")
            case _:
                pass

        return reactions[0]

    @property
    def dict(self) -> dict:
        """Returns a dictionary of this compound. See :meth:`.Compound.get_dict`"""
        return self.get_dict()

    @property
    def is_base(self) -> bool:
        """Is this Compound the basis for any elaborations?"""
        return bool(
            self.db.select_where(
                query="1",
                table="scaffold",
                key="base",
                value=self.id,
                multiple=False,
                none="quiet",
            )
        )

    @property
    def is_elab(self) -> bool:
        """Is this Compound the based on any other compound?"""
        return bool(
            self.db.select_where(
                query="1",
                table="scaffold",
                key="superstructure",
                value=self.id,
                multiple=False,
                none="quiet",
            )
        )

    @property
    def is_product(self) -> bool:
        """Is this Compound a product of at least one reaction"""
        return bool(self.get_reactions(none=False))

    @property
    def table(self):
        """Returns the name of the :class:`.Database` table"""
        return self._table

    @property
    def _db_changed(self) -> bool:
        """Has the database changed?"""
        if self._total_changes != self.db.total_changes:
            self._total_changes = self.db.total_changes
            return True
        return False

    ### METHODS

    def add_stock(
        self,
        amount: float,
        *,
        purity: float | None = None,
        entry: str | None = None,
        location: str | None = None,
        return_quote: bool = True,
    ) -> int | Quote:
        """Register a certain quantity of compound stock in the Database.

        :param amount: Amount in ``mg``
        :param purity: Purity fraction ``0 < purity <= 1``, defaults to ``None``
        :param location: String describing where this stock is located, defaults to ``None``
        :param return_quote: If ``True`` a :class:`.Quote` object is returned instead of its ID, defaults to ``True``
        :returns: The inserted :class:`.Quote` object or ID (see ``return_quote``)
        """

        assert amount

        # search for existing in stock quotes
        existing = self.get_quotes(supplier="Stock", df=False)

        # supersede old in stock records
        if existing:
            delete = set()
            not_deleted = 0
            for quote in existing:

                if any(
                    [
                        quote.entry != entry,
                        quote.purity != purity,
                        quote.catalogue != location,
                    ]
                ):
                    not_deleted += 1
                    continue

                delete.add(quote.id)

            delete_str = str(tuple(delete)).replace(",)", ")")

            self.db.delete_where(table="quote", key=f"quote_id IN {delete_str}")

            if delete:
                mrich.warning(f"Removed {len(delete)} existing In-Stock Quotes")

            if not_deleted:
                mrich.warning(
                    f"Did not remove {not_deleted} existing In-Stock Quotes with differing entry/purity/location"
                )

        # insert the new quote
        quote_id = self.db.insert_quote(
            compound=self.id,
            price=0,
            lead_time=0,
            currency=None,
            supplier="Stock",
            catalogue=location,
            entry=entry,
            amount=amount,
            purity=purity,
        )

        if return_quote:
            self.db.get_quote(id=quote_id)
        else:
            return quote_id

    def get_tags(self) -> "TagSet":
        """Get the tags assigned to this compound"""
        tags = self.db.select_where(
            query="tag_name",
            table="tag",
            key="compound",
            value=self.id,
            multiple=True,
            none="quiet",
        )
        return TagSet(self, {t[0] for t in tags}, commit=False)

    def add_tag(
        self,
        tag: str,
    ) -> None:
        """Add this tag to every member of the set"""
        assert isinstance(tag, str)
        self.db.insert_tag(name=tag, compound=self.id, commit=True)

    def get_quotes(
        self,
        min_amount: float | None = None,
        supplier: str | None = None,
        max_lead_time: float | None = None,
        none: str = "quiet",
        pick_cheapest: bool = False,
        df: bool = False,
    ) -> list["Quote"]:
        """Get all quotes associated to this compound

        :param min_amount: Only return quotes with amounts greater than this, defaults to ``None``
        :param supplier: Only return quotes with the given supplier, defaults to ``None``
        :param max_lead_time: Only return quotes with lead times less than this (in days), defaults to ``None``
        :param none: Define the behaviour when no quotes are found. Choose `error` to raise print an error.
        :param pick_cheapest: If ``True`` only the cheapest :class:`.Quote` is returned, defaults to ``False``
        :param df: Returns a ``DataFrame`` of the quoting data, defaults to ``False``
        :returns: List of :class:`.Quote` objects, ``DataFrame``, or single :class:`.Quote`. See ``pick_cheapest`` and ``df`` parameters

        """

        if not supplier:
            quote_ids = self.db.select_where(
                query="quote_id",
                table="quote",
                key="compound",
                value=self.id,
                multiple=True,
                none=none,
            )
        elif isinstance(supplier, str):
            quote_ids = self.db.select_where(
                query="quote_id",
                table="quote",
                key=f'quote_compound = {self.id} AND quote_supplier = "{supplier}"',
                multiple=True,
                none=none,
            )
        else:
            quote_ids = self.db.select_where(
                query="quote_id",
                table="quote",
                key=f'quote_compound = {self.id} AND quote_supplier IN {str(tuple(supplier)).replace(",)",")")}',
                multiple=True,
                none=none,
            )

        if quote_ids:
            quotes = [self.db.get_quote(id=q[0]) for q in quote_ids]
        else:
            return None

        if max_lead_time:
            quotes = [q for q in quotes if q.lead_time <= max_lead_time]

        if min_amount:
            suitable_quotes = [q for q in quotes if q.amount >= min_amount]

            if not suitable_quotes:
                mrich.debug(f"No quote available with amount >= {min_amount} mg")
                quotes = [Quote.combination(min_amount, quotes)]

            else:
                quotes = suitable_quotes

        if pick_cheapest:
            return sorted(quotes, key=lambda x: x.price)[0]

        if df:
            from pandas import DataFrame

            return DataFrame([q.dict for q in quotes]).drop(columns="compound")

        return quotes

    def get_reactions(
        self,
        as_reactant: bool = False,
        permitted_reactions: "ReactionSet" = None,
        none: str = "error",
    ) -> "ReactionSet":
        """Get the associated :class:`.Reaction` objects. By default this function returns all reaction resulting in this :class:`.Compound` as a product, unless ``as_reactant`` is set to ``True``.

        :param as_reactant: Search for :class:`.Reaction` objects using this :class:`.Compound` as a reactant instead of a product, defaults to ``False``
        :param permitted_reactions: Provide a :class:`.ReactionSet` by which to filter the results
        :param none: Define the behaviour when no quotes are found. Choose `error` to raise print an error, defaults to ``'error'``
        """

        from .rset import ReactionSet

        if as_reactant:
            reaction_ids = self.db.select_where(
                query="reactant_reaction",
                table="reactant",
                key="compound",
                value=self.id,
                multiple=True,
                none=none,
            )
        else:
            reaction_ids = self.db.select_where(
                query="reaction_id",
                table="reaction",
                key="product",
                value=self.id,
                multiple=True,
                none=none,
            )

        reaction_ids = [q for q, in reaction_ids]

        if permitted_reactions:
            reaction_ids = [i for i in reaction_ids if i in permitted_reactions]

        rset = ReactionSet(self.db, reaction_ids)

        if not permitted_reactions:
            rset._name = f"reactions resulting in {str(self)}"

        return rset

    def get_poses(self) -> "PoseSet":
        """Get the associated :class:`.Pose` objects."""

        pose_ids = self.db.select_where(
            query="pose_id",
            table="pose",
            key="compound",
            value=self.id,
            multiple=True,
            none=False,
        )

        from .pset import PoseSet

        return PoseSet(self.db, [q[0] for q in pose_ids], name=f"{self}'s poses")

    def get_dict(
        self,
        *,
        mol: bool = True,
        alias: bool = True,
        inchikey: bool = True,
        metadata: bool = True,
        poses: bool = True,
        count_by_target: bool = False,
        num_reactant: bool = True,
        num_reactions: bool = True,
        bases: bool = True,
        elabs: bool = True,
        tags: bool = True,
    ) -> "dict":
        """Returns a dictionary representing this :class:`.Compound`

        :param mol: Include a ``rdkit.Chem.Mol object``, defaults to ``True``
        :param metadata: Include metadata, defaults to ``True``
        :param poses: Include dictionaries of associated :class:`.Pose` objects, defaults to ``True``
        :param count_by_target: Include counts by protein :class:`.Target`, defaults to ``False``. Only applicable when ``count_by_target = True``.
        :param num_reactant: include num_reactant column
        :param num_reactions: include num_reactions column
        :param bases: include bases column
        :param elabs: include elabs column
        :param tags: include tags column
        :returns: A dictionary
        """

        serialisable_fields = [
            "id",
            "smiles",
        ]

        if alias:
            serialisable_fields.append("alias")
        if inchikey:
            serialisable_fields.append("inchikey")
        if num_reactant:
            serialisable_fields.append("num_reactant")
        if num_reactions:
            serialisable_fields.append("num_reactions")

        data = {}
        for key in serialisable_fields:
            data[key] = getattr(self, key)

        if mol:
            try:
                data["mol"] = self.mol
            except InvalidMolError:
                data["mol"] = None

        if bases:
            if self.bases:
                data["bases"] = self.bases.ids
            else:
                data["bases"] = None

        if elabs:
            if self.elabs:
                data["elabs"] = self.elabs.ids
            else:
                data["elabs"] = None

        if tags:
            data["tags"] = self.tags

        if poses:

            poses = self.poses

            if poses:

                data["poses"] = poses.ids
                data["targets"] = poses.target_names

                if count_by_target:
                    target_ids = poses.target_ids

                    for target in self._animal.targets:
                        t_poses = poses(target=target.id) or []
                        data[f"#poses {target.name}"] = len(t_poses)

        if metadata and (metadict := self.metadata):
            for key in metadict:
                data[key] = metadict[key]

        return data

    def get_recipes(
        self,
        *,
        amount: float = 1,
        debug: bool = False,
        pick_cheapest: bool = False,
        quoted_only: bool = False,
        supplier: None | str = None,
        **kwargs,
    ):
        """Get :class:`.Recipe` objects that result in this compound. See :meth:`.Recipe.from_compounds`"""

        from .recipe import Recipe
        from .cset import CompoundSet

        return Recipe.from_compounds(
            CompoundSet(self.db, [self.id]),
            amount=amount,
            debug=debug,
            pick_cheapest=pick_cheapest,
            quoted_only=quoted_only,
            supplier=supplier,
            **kwargs,
        )

    def get_base_ids(self) -> list[int]:
        """Get a list of :class:`.Compound` ID's that this object is a superstructure of"""
        ids = self.db.select_where(
            table="scaffold",
            query="scaffold_base",
            key="superstructure",
            value=self.id,
            none="quiet",
            multiple=True,
        )
        if not ids:
            return None
        return [i for i, in ids]

    def get_superstructure_ids(self) -> list[int]:
        """Get a list of :class:`.Compound` ID's that this object is a substructure of"""
        ids = self.db.select_where(
            table="scaffold",
            query="scaffold_superstructure",
            key="base",
            value=self.id,
            none="quiet",
            multiple=True,
        )
        if not ids:
            return None
        return [i for i, in ids]

    def add_base(self, base: "Compound | int", commit: bool = True) -> None:
        """
        Add a base :class:`.Compound` this molecule is derived from.

        :param base: The base :class:`.Compound` or its ID.
        :param commit: Commit the changes to the :class:`.Database`, defaults to ``True``
        """

        if not isinstance(base, int):
            assert base._table == "compound"
            base = base.id
        self.db.insert_scaffold(base=base, superstructure=self.id)

    def set_alias(self, alias: str, commit=True) -> None:
        """
        Set this :class:`.Compound`'s alias.

        :param alias: The alias
        :param commit: Commit the changes to the :class:`.Database`, defaults to ``True``
        """

        assert isinstance(alias, str)
        self._alias = alias
        self.db.update(
            table="compound",
            id=self.id,
            key="compound_alias",
            value=alias,
            commit=commit,
        )

    def as_ingredient(
        self,
        amount: float,
        max_lead_time: float = None,
        supplier: str = None,
        get_quote: bool = True,
        quote_none: str = "quiet",
    ) -> "Ingredient":
        """Convert this compound into an :class:`.Ingredient` object with an associated amount (in ``mg``) and :class:`.Quote` if available.

        :param amount: Amount in ``mg``
        :param supplier: Only search for quotes with the given supplier, defaults to ``None``
        :param max_lead_time: Only search for quotes with lead times less than this (in days), defaults to ``None``
        """

        if get_quote:
            quote = self.get_quotes(
                pick_cheapest=True,
                min_amount=amount,
                max_lead_time=max_lead_time,
                supplier=supplier,
                none=quote_none,
            )

            if not quote:
                quote = None

        else:
            quote = None

        return Ingredient(
            db=self.db,
            compound=self.id,
            amount=amount,
            quote=quote,
            supplier=supplier,
            max_lead_time=max_lead_time,
        )

    def draw(self, align_substructure: bool = False) -> None:
        """Display this compound (and its base if it has one)

        .. attention::

                This method is only intended for use within a Jupyter Notebook.

        :param align_substructure: Align the two drawing by their common substructure, defaults to ``False``
        """

        if bases := self.bases:

            from molparse.rdkit import draw_mcs

            data = {}
            for base in bases:
                data[base.smiles] = f"{base} (base)"
            data[self.smiles] = str(self)

            if len(data) > 1:

                drawing = draw_mcs(
                    data,
                    align_substructure=align_substructure,
                    show_mcs=False,
                    highlight_common=False,
                )
                display(drawing)

            else:
                mrich.error(
                    f"Problem drawing {base.id=} vs {self.id=}, self referential?"
                )
                display(self.mol)

        else:
            display(self.mol)

    def draw_elabs(self):

        from molparse.rdkit import draw_highlighted_mol
        from rdkit.Chem import rdRGroupDecomposition, MolFromSmarts

        elabs = self.elabs

        display(self)
        display(elabs)

        if not elabs:
            mrich.error(self, "has no elaborations")
            return self.draw()

        # set RGD params
        params = rdRGroupDecomposition.RGroupDecompositionParameters()
        params.removeAllHydrogenRGroups = False
        params.removeAllHydrogenRGroupsAndLabels = True
        params.removeHydrogensPostMatch = True

        # do the RGD
        rgd = rdRGroupDecomposition.RGroupDecomposition(
            MolFromSmarts(self.smiles), params
        )
        for mol in elabs.mols:
            rgd.Add(mol)
        rgd.Process()

        # Get the R-group decomposition results
        rgroup_table = rgd.GetRGroupsAsColumns()

        # get the core and its attachment points
        core = rgroup_table["Core"][0]
        attachment_points = set()
        for rgroup in rgroup_table["Core"]:
            for atom in rgroup.GetAtoms():
                if atom.GetAtomicNum() == 0:  # Dummy atom (R-group attachment point)
                    attachment_points.add(atom.GetIdx())

        # display the annotated core
        drawing = draw_highlighted_mol(
            core, [(i, (0.5, 1, 0.5)) for i in attachment_points]
        )
        display(drawing)

    def classify(
        self,
        draw: bool = True,
    ) -> list[tuple[str, int]]:
        """
        Find RDKit Fragments within the compound molecule and draw them

        :param draw: Draw the annotated molecule, defaults to ``True``
        :returns: A list of tuples containing a descriptor (``str``) and count (``int``) pair
        """
        # from molparse.rdkit import classify_mol
        from molparse.rdkit.classify import classify_mol

        return classify_mol(self.mol, draw=draw)

    def murcko_scaffold(
        self,
        generic: bool = False,
    ):
        """Get the rdkit MurckoScaffold for this compound"""

        from rdkit.Chem.Scaffolds import MurckoScaffold

        scaffold = MurckoScaffold.GetScaffoldForMol(self.mol)

        if generic:
            scaffold = MurckoScaffold.MakeScaffoldGeneric(scaffold)

        return scaffold

    def summary(
        self, metadata: bool = True, draw: bool = True, tags: bool = True
    ) -> None:
        """
        Print a summary of this compound

        :param metadata: Include metadata, defaults to ``True``
        :param draw: Include a 2D molecule drawing, defaults to ``True``
        """

        mrich.header(self)

        mrich.var("inchikey", self.inchikey)
        mrich.var("alias", self.alias)
        mrich.var("smiles", self.smiles)
        mrich.var("bases", self.bases)
        mrich.var("elabs", self.elabs)

        mrich.var("is_base", self.is_base)
        mrich.var("is_elab", self.is_elab)
        mrich.var("num_heavy_atoms", self.num_heavy_atoms)
        mrich.var("num_rings", self.num_rings)
        mrich.var("formula", self.formula)

        mrich.var("#reactions (product)", self.num_reactions)
        mrich.var("#reactions (reactant)", self.num_reactant)

        if tags:
            mrich.var("tags", self.tags)

        poses = self.poses
        mrich.var("#poses", len(poses))
        if poses:
            mrich.var("targets", poses.targets)

        if metadata:
            mrich.var("metadata", str(self.metadata))

        if draw:
            self.draw()

    def place(
        self,
        *,
        reference: Pose,
        inspirations: list[Pose] | None = None,
        max_ddG: float = 0.0,
        max_RMSD: float = 2.0,
        output_dir: str = "wictor_place",
        tags: list[str] = None,
        metadata: dict = None,
        overwrite: bool = False,
    ) -> Pose:
        """
        Generate a new pose for this compound using Fragmenstein.

        :param reference: Choose the :class:`.Pose` to use as the reference protein conformation
        :param inspirations: Choose the (virtual) hits to to define the ligand reference, defaults to the ``reference``'s inspirations
        :param max_ddG: Maximum ``ddG`` value permitted for a valid ligand conformation, defaults to ``0.0``
        :param max_RMSD: Maximum ``RMSD`` value permitted for a valid ligand conformation, defaults to ``2.0``
        :param output_dir: Output directory for Fragmenstein files, defaults to ``wictor_place``
        :param tags: Tags to assign to the created pose, defaults to ``[]``
        :param metadata: A dictionary of metadata to assign to this compound, defaults to ``{}``
        :param overwrite: Delete old poses, defaults to ``False``
        """

        from fragmenstein import Monster, Wictor
        from pathlib import Path

        tags = tags or []
        metadata = metadata or {}

        # get required data
        smiles = self.smiles

        inspirations = inspirations or reference.inspirations
        target = reference.target.name

        inspiration_mols = [c.mol for c in inspirations]
        protein_pdb_block = reference.protein_system.pdb_block_with_alt_sites

        # create the victor
        victor = Wictor(hits=inspiration_mols, pdb_block=protein_pdb_block)
        victor.work_path = output_dir
        victor.enable_stdout(logging.CRITICAL)

        # do the placement
        victor.place(smiles, long_name=self.name)

        # metadata
        metadata["ddG"] = (
            victor.energy_score["bound"]["total_score"]
            - victor.energy_score["unbound"]["total_score"]
        )
        metadata["RMSD"] = victor.mrmsd.mrmsd

        if metadata["ddG"] > max_ddG:
            return None

        if metadata["RMSD"] > max_RMSD:
            return None

        # register the pose
        pose = self._animal.register_pose(
            compound=self,
            target=target,
            path=Path(victor.work_path) / self.name / f"{self.name}.minimised.mol",
            inspirations=inspirations,
            reference=reference,
            tags=tags,
            metadata=metadata,
        )

        if overwrite:
            ids = [p.id for p in self.poses if p.id != pose.id]
            for i in ids:
                self.db.delete_where(table="pose", key="id", value=i)
            mrich.success(f"Successfully posed {self} (and deleted old poses)")
        else:
            mrich.success(f"Successfully posed {self}")

        return pose

    def get_inspirations(self, debug: bool = True, none: str = "warning") -> "PoseSet":
        """Since inspirations map :class:`.Pose` objects to each other rather than :class:`.Compound` objects, this only works if there are poses registerd for this compound or it's elaborations/superstructures.

        :returns: a :class:`.PoseSet` object
        """

        from .pset import PoseSet

        sql = """
        SELECT pose_id, inspiration_original FROM compound
        INNER JOIN scaffold ON compound_id = scaffold_base
        INNER JOIN pose ON compound_id = pose_compound
        INNER JOIN inspiration ON pose_id = inspiration_derivative
        WHERE compound_id = :compound_id
        """

        with mrich.spinner(f"Querying inspirations for {self}"):
            records = self.db.execute(sql, dict(compound_id=self.id)).fetchall()

        if not records and none in ("warning", "warn"):
            mrich.warning("Could not determine inspirations for", self)
            return None

        derivatives = PoseSet(self.db, set(a for a, b in records))
        inspirations = PoseSet(self.db, set(b for a, b in records))

        if debug:
            mrich.debug(f"Inspirations derived from {derivatives.ids}")

        inspirations._name = f"Inspirations for {self}"

        return inspirations

    ### DUNDERS

    def __str__(self) -> str:
        """Unformatted string representation"""
        return f"C{self.id}"

    def __repr__(self) -> str:
        """ANSI Formatted string representation"""
        return f'{mcol.bold}{mcol.underline}{self} "{self.name}"{mcol.unbold}{mcol.ununderline}'

    def __rich__(self) -> str:
        """Representation for mrich"""
        return f'[bold underline]{self} "{self.name}"'

    def __eq__(self, other) -> bool:
        """Compare compounds"""
        assert isinstance(other, Compound)
        return self.id == other.id


class Ingredient:
    """An ingredient is a :class:`.Compound` with a fixed quanitity and an attached quote.

    .. image:: ../images/ingredient.png
              :width: 450
              :alt: Ingredient schema

    .. attention::

            :class:`.Ingredient` objects should not be created directly. Instead use :meth:`.Compound.as_ingredient`.
    """

    _table = "ingredient"

    def __init__(
        self, db, compound, amount, quote, max_lead_time=None, supplier=None
    ) -> "Ingredient":

        assert compound

        self._db = db

        # don't store inherited compound in memory until needed
        self._compound = None

        if isinstance(compound, Compound):
            self._compound_id = compound.id
            self._compound = None
        else:
            self._compound_id = compound

        if isinstance(quote, Quote):

            if id := quote.id:
                self._quote_id = quote.id
                self._quote = None

            else:
                self._quote_id = None
                self._quote = quote

        elif quote is None:
            self._quote_id = None
            self._quote = None

        else:
            self._quote_id = int(quote)
            self._quote = None

        self._amount = amount
        self._max_lead_time = max_lead_time
        self._supplier = supplier
        self._total_changes = db.total_changes

    ### PROPERTIES

    @property
    def db(self) -> "Database":
        """Returns the parent :class:`.Database`"""
        return self._db

    @property
    def amount(self) -> float:
        """Returns the amount (in ``mg``)"""
        return self._amount

    @property
    def id(self) -> int:
        """Returns the ID of the associated :class:`.Compound`"""
        return self._compound_id

    @property
    def compound_id(self) -> int:
        """Returns the ID of the associated :class:`.Compound`"""
        return self._compound_id

    @property
    def quote_id(self) -> int:
        """Returns the ID of the associated :class:`.Quote`"""
        return self._quote_id

    @property
    def max_lead_time(self) -> float:
        """Returns the max_lead_time (in days) from the original quote query"""
        return self._max_lead_time

    @property
    def supplier(self) -> str:
        """Returns the supplier from the original quote query"""
        return self._supplier

    @amount.setter
    def amount(self, a) -> None:
        """Set the amount and fetch updated :class:`.Quote`s"""

        quote_id = self.get_cheapest_quote_id(
            min_amount=a,
            max_lead_time=self._max_lead_time,
            supplier=self._supplier,
            none="quiet",
        )

        self._quote_id = quote_id

        self._amount = a

    @property
    def compound(self) -> Compound:
        """Returns the associated :class:`.Compound`"""

        if not self._compound:
            self._compound = self.db.get_compound(id=self.compound_id)
        return self._compound

    @property
    def quote(self) -> Quote:
        """Returns the associated :class:`.Quote`"""

        if self._quote is None:

            if q_id := self.quote_id:
                self._quote = self.db.get_quote(id=self.quote_id)

            else:
                q = self.compound.get_quotes(
                    pick_cheapest=True,
                    min_amount=self.amount,
                    max_lead_time=self.max_lead_time,
                    supplier=self.supplier,
                    none="quiet",
                )

                if not q:
                    return None

                self._quote = q
                self._quote_id = q.id

        return self._quote

    @property
    def compound_price_amount_str(self) -> str:
        """String representation including :class:`.Compound`, :class:`.Price`, and amount."""
        return f"{self} ({self.amount})"

    @property
    def smiles(self) -> str:
        """Returns the SMILES of the associated :class:`.Compound`"""
        return self.compound.smiles

    @property
    def price(self) -> "Price | None":
        """Returns the :class:`.Price` of the associated :class:`.Quote`"""
        if self.quote:
            return self.quote.price
        else:
            return None

    @property
    def lead_time(self) -> float | None:
        """Returns the lead time (in days) of the associated :class:`.Quote`"""
        if self.quote:
            return self.quote.lead_time
        else:
            return None

    @property
    def _db_changed(self) -> bool:
        """Has the database changed?"""
        if self._total_changes != self.db.total_changes:
            self._total_changes = self.db.total_changes
            return True
        return False

    ### METHODS

    def get_cheapest_quote_id(
        self,
        min_amount: float | None = None,
        supplier: str | None = None,
        max_lead_time: float | None = None,
        none: str = "quiet",
    ) -> int | None:
        """
        Query quotes associated to this ingredient, and return the cheapest

        :param min_amount: Only return quotes with amounts greater than this, defaults to ``None``
        :param supplier: Only return quotes with the given supplier, defaults to ``None``
        :param max_lead_time: Only return quotes with lead times less than this (in days), defaults to ``None``
        :param none: Define the behaviour when no quotes are found. Choose `error` to raise print an error.
        """

        supplier_str = f' AND quote_supplier IS "{supplier}"' if supplier else ""
        lead_time_str = (
            f" AND quote_lead_time <= {max_lead_time}" if max_lead_time else ""
        )
        key_str = f"quote_compound IS {self.compound_id} AND quote_amount >= {min_amount}{supplier_str}{lead_time_str} ORDER BY quote_price"

        result = self.db.select_where(
            query="quote_id", table="quote", key=key_str, multiple=False, none=none
        )

        if result:
            (quote_id,) = result
            return quote_id

        else:
            return None

    def get_quotes(self, **kwargs) -> list["Quote"]:
        """Wrapper for :meth:`.Compound.get_quotes()`"""
        return self.compound.get_quotes(**kwargs)

    ### DUNDERS

    def __str__(self) -> str:
        """Plain string representation"""
        return f"{self.amount:.2f}mg of C{self._compound_id}"

    def __repr__(self) -> str:
        """ANSI Formatted string representation"""
        return f"{mcol.bold}{mcol.underline}{str(self)}{mcol.unbold}{mcol.ununderline}"

    def __rich__(self) -> str:
        """Representation for mrich"""
        return f"[bold underline]{str(self)}"

    def __eq__(self, other) -> bool:
        """Equality operator"""

        if self.compound_id != other.compound_id:
            return False

        return self.amount == other.amount

    def __getattr__(self, key: str):
        """For missing attributes try getting from associated :class:`.Compound`"""
        return getattr(self.compound, key)



import mrich


def get_scaffold_network(
    animal,
    compounds="CompoundSet | None",
    scaffolds="CompoundSet | None",
    # filename: "str | Path" = "network.html",
    notebook: bool = True,
    depth: int = 5,
    scaffold_tag: str | None = None,
    exclude_tag: str | None = None,
    physics: bool = True,
    arrows: bool = True,
) -> "pyvis.network.Network":
    """Use PyVis to display a network of molecules connected by scaffold relationships in the database"""

    from pyvis.network import Network
    from molparse.rdkit import smiles_to_pngstr

    net = Network(notebook=notebook, cdn_resources="in_line")

    nodes = set()
    edges = set()

    arrows = "to" if arrows else None

    def add_node(compound):

        if compound.id in nodes:
            return

        pngstr = smiles_to_pngstr(compound.smiles)

        net.add_node(
            compound.id,
            label=compound.alias or str(compound),
            title=str(compound),
            shape="circularImage",
            image=f"data:image/png;base64,{pngstr}",
            physics=physics,
        )

        nodes.add(compound.id)

    def add_edge(base, compound):

        key = (base.id, compound.id)

        if key in edges:
            return

        net.add_edge(
            base.id,
            compound.id,
            arrows=arrows,
            # label="PDE5",
            # title="Protein target",
            # color="purple"
        )

        edges.add(key)

    def get_scaffold_records(scaffolds=None, compounds=None):
        if scaffolds:
            return animal.db.select_all_where(
                table="scaffold",
                key=f"scaffold_base IN {scaffolds.str_ids}",
                multiple=True,
                none="quiet",
            )
        elif compounds:
            return animal.db.select_all_where(
                table="scaffold",
                key=f"scaffold_superstructure IN {compounds.str_ids}",
                multiple=True,
                none="quiet",
            )
        raise ValueError

    if compounds and not scaffolds:

        mrich.var("recursion depth", depth)
        mrich.var("#compounds", len(compounds))

        records = []
        n = 0
        while n < depth:

            if not compounds:
                break

            results = get_scaffold_records(compounds=compounds)

            if results:
                records.extend(results)
                compounds = animal.compounds[[a for a, b in results]]
                n += 1
            else:
                break

        if n == depth:
            mrich.warning(
                "Reached recursion depth. More scaffolds may be in the database"
            )

    elif scaffolds and not compounds:

        mrich.var("recursion depth", depth)
        mrich.var("#scaffolds", len(scaffolds))

        records = []
        n = 0
        while n < depth:

            if not scaffolds:
                break

            results = get_scaffold_records(scaffolds=scaffolds)

            if results:
                records.extend(results)
                scaffolds = animal.compounds[[b for a, b in results]]
                n += 1
            else:
                break

        if n == depth:
            mrich.warning(
                "Reached recursion depth. More superstructures may be in the database"
            )

    else:
        raise ValueError

    mrich.var("#edges", len(records))

    if records:
        for base_id, compound_id in mrich.track(
            records, prefix="Adding nodes and edges"
        ):

            base = animal.db.get_compound(id=base_id)

            base_tags = base.tags

            if scaffold_tag and scaffold_tag not in base_tags:
                continue

            compound = animal.db.get_compound(id=compound_id)

            if exclude_tag and (
                exclude_tag in base_tags or exclude_tag in compound.tags
            ):
                continue

            add_node(base)
            add_node(compound)
            add_edge(base, compound)

    return net



import re
import numpy as np
from molparse.rdkit import mol_from_smiles
from rdkit.Chem.inchi import MolToInchiKey
from rdkit.Chem import MolFromSmiles, MolToSmiles, AddHs, RemoveHs
import mcol
import mout
from datetime import datetime
from string import ascii_uppercase

import mrich


def df_row_to_dict(df_row):
    """

    :param df_row:

    """

    assert len(df_row) == 1, f"{len(df_row)=}"

    data = {}

    for col in df_row.columns:

        if col == "Unnamed: 0":
            continue

        value = df_row[col].values[0]

        if not isinstance(value, str) and np.isnan(value):
            value = None

        data[col] = value

    return data


def remove_other_ligands(sys, residue_number, chain):
    """

    :param sys:
    :param residue_number:
    :param chain:

    """

    ligand_residues = [r.number for r in sys["rLIG"] if r.number != residue_number]

    # if ligand_residues:
    for c in sys.chains:
        if c.name != chain:
            c.remove_residues(names=["LIG"], verbosity=0)
        elif ligand_residues:
            c.remove_residues(numbers=ligand_residues, verbosity=0)

    # print([r.name_number_str for r in sys['rLIG']])

    assert (
        len([r.name_number_str for r in sys["rLIG"]]) == 1
    ), f"{sys.name} {[r.name_number_str for r in sys['rLIG']]}"

    return sys


def inchikey_from_smiles(smiles):
    """

    :param smiles:

    """
    mol = mol_from_smiles(smiles)
    return MolToInchiKey(mol)


def flat_inchikey(smiles):
    """

    :param smiles:

    """
    smiles = sanitise_smiles(smiles)
    return inchikey_from_smiles(smiles)


def remove_isotopes_from_smiles(smiles):
    """

    :param smiles:

    """

    mol = MolFromSmiles(smiles)

    atom_data = [(atom, atom.GetIsotope()) for atom in mol.GetAtoms()]

    for atom, isotope in atom_data:
        if isotope:
            atom.SetIsotope(0)

    return MolToSmiles(mol)


def smiles_has_isotope(smiles, regex=True):
    """

    :param smiles:
    :param regex:  (Default value = True)

    """
    if regex:
        return re.search(r"([\[][0-9]+[A-Z]+\])", smiles)
    else:
        mol = MolFromSmiles(smiles)
        return any(atom.GetIsotope() for atom in mol.GetAtoms())


def sanitise_smiles(s, verbosity=False, sanitisation_failed="error", radical="error"):
    """

    :param s:
    :param verbosity:  (Default value = False)
    :param sanitisation_failed:  (Default value = 'error')
    :param radical:  (Default value = 'error')

    """

    assert isinstance(s, str), f"non-string smiles={s}"

    orig_smiles = s

    # if multiple molecules take the largest
    if "." in s:
        s = sorted(s.split("."), key=lambda x: len(x))[-1]

    # flatten the smiles
    stereo_smiles = s
    smiles = s.replace("@", "")
    smiles = smiles.replace("/", "")
    smiles = smiles.replace("\\", "")

    # remove isotopic stuff
    if smiles_has_isotope(smiles):
        mout.warning(f"Isotope(s) in SMILES: {smiles}")
        smiles = remove_isotopes_from_smiles(smiles)

    # canonicalise
    mol = MolFromSmiles(smiles)
    if mol:
        smiles = MolToSmiles(mol, True)
    elif sanitisation_failed == "error":
        raise SanitisationError
    elif sanitisation_failed == "warning":
        mrich.warning(f"sanitisation failed for {smiles=}")

    # check radicals
    reconstruct = False
    for atom in mol.GetAtoms():
        if not atom.GetNumRadicalElectrons():
            continue

        if radical == "warning":
            mrich.warning(f"Radical atom in {smiles=}")
        elif radical == "error":
            raise SanitisationError(f"Radical atom in {smiles=}")
        elif radical == "remove":
            mrich.warning(f"Removed radical atom")
            atom.SetNumRadicalElectrons(0)
            smiles = MolToSmiles(mol, True)
            reconstruct = True
            # atom.SetFormalCharge(0)
        else:
            raise NotImplementedError(f"Unknown option {radical=}")

    if reconstruct:
        mol = AddHs(mol)
        mol = RemoveHs(mol, implicitOnly=True)
        smiles = MolToSmiles(mol, True)
        mrich.warning(f"New {smiles=}")

    if verbosity:

        if smiles != orig_smiles:

            annotated_smiles_str = orig_smiles.replace(
                ".", f"{mcol.error}{mcol.underline}.{mcol.clear}{mcol.warning}"
            )
            annotated_smiles_str = annotated_smiles_str.replace(
                "@", f"{mcol.error}{mcol.underline}@{mcol.clear}{mcol.warning}"
            )

            mout.warning(f"SMILES was changed: {annotated_smiles_str} --> {smiles}")

    return smiles


def sanitise_mol(m):
    """

    :param m:

    """
    from rdkit.Chem import MolToMolBlock, MolFromMolBlock

    return MolFromMolBlock(MolToMolBlock(m))


def pose_gap(a, b):
    """

    :param a:
    :param b:

    """

    from numpy.linalg import norm
    from molparse.rdkit import mol_to_AtomGroup

    min_dist = None

    a = mol_to_AtomGroup(a.mol)
    b = mol_to_AtomGroup(b.mol)

    for atom1 in a.atoms:
        for atom2 in b.atoms:
            dist = norm(atom1.np_pos - atom2.np_pos)
            if min_dist is None or dist < min_dist:
                min_dist = dist

    return min_dist


ALPHANUMERIC_CHARS = "0123456789" + ascii_uppercase


def numberToBase(n, b):
    if n == 0:
        return [0]
    digits = []
    while n:
        digits.append(int(n % b))
        n //= b
    return digits[::-1]


def dt_hash():
    dt = datetime.now()
    x = int(
        dt.month * 36000 * 24 * 365.25
        + dt.day * 36000 * 24
        + dt.hour * 36000
        + dt.minute * 600
        + dt.second * 10
        + dt.microsecond / 10000
    )
    timehash = "".join([ALPHANUMERIC_CHARS[v] for v in numberToBase(x, 36)])
    return f"{timehash:>07}"


class SanitisationError(Exception):
    """ """

    ...



from .animal import HIPPO
import mrich
from typer import Typer
from pathlib import Path

app = Typer()


def setup_animal(
    database: str,
    backup: bool = True,
    update_legacy: bool = False,
) -> HIPPO:
    animal = HIPPO("CLI", database, update_legacy=update_legacy)
    if backup:
        animal.db.backup()
    return animal


@app.command()
def backup(database: str):
    """Backup database file"""
    mrich.h1("hippo.backup")

    mrich.h3("Params")
    mrich.var("database", database)

    from hippo.db import backup

    backup(database)
    mrich.success("Successfully backed up")


@app.command()
def update_legacy(database: str, backup: bool = True):
    """Update legacy database format"""

    mrich.h1("hippo.update_legacy")

    mrich.h3("Params")
    mrich.var("database", database)

    if backup:
        from hippo.db import backup

        backup(database)

    animal = setup_animal(database, backup=False, update_legacy=True)
    mrich.success("Successfully updated database format")


@app.command()
def calculate_scaffolds(
    database: str,
    backup: bool = True,
):
    """Calculate scaffold/superstructure relationships for all compounds"""

    mrich.h1("hippo.calculate_scaffolds")

    mrich.h3("Params")
    mrich.var("database", database)
    mrich.var("backup", backup)

    mrich.h3("Animal")
    animal = setup_animal(database=database, backup=backup)

    mrich.h3("State Before")
    mrich.var("bases", animal.bases)
    mrich.var("elabs", animal.elabs)

    mrich.h3("Calculation")
    animal.db.calculate_all_scaffolds()

    mrich.h3("State After")
    mrich.var("bases", animal.bases)
    mrich.var("elabs", animal.elabs)

    mrich.success("Completed")


@app.command()
def calculate_interactions(
    database: str,
    prolif: bool = False,
    backup: bool = True,
    force: bool = False,
):
    """Calculate interactions for all poses"""

    mrich.h1("hippo.calculate_interactions")

    mrich.h3("Params")
    mrich.var("database", database)
    mrich.var("backup", backup)

    mrich.h3("Animal")
    animal = setup_animal(database=database, backup=backup)

    mrich.h3("State Before")
    mrich.var("#fingerprinted", animal.poses.num_fingerprinted)

    mrich.h3("Calculation")

    n = len(animal.poses)
    for i, pose in mrich.track(enumerate(animal.poses), total=n):

        mrich.set_progress_prefix(f"{i}/{n}")

        try:
            if prolif:
                pose.calculate_prolif_interactions(force=force)
            else:
                pose.calculate_interactions(force=force)

        except Exception as e:
            mrich.error(e)
            mrich.error("Could not fingerprint pose")
            continue

    mrich.h3("State After")
    mrich.var("#fingerprinted", animal.poses.num_fingerprinted)

    mrich.success("Completed")


@app.command()
def verify():
    """Verify installation"""

    import os

    file_path = "_test.sqlite"

    try:
        animal = setup_animal(file_path, backup=False)
        c = animal.register_compound(smiles="COc1ccc2sc(N)nc2c1")
        c.mol
        mrich.success("HIPPO/rdkit/chemicalite installations are compatible")
    except Exception as e:
        mrich.error(e)

    if os.path.exists(file_path):
        os.remove(file_path)


@app.command()
def tag_summary(
    database: str,
):
    """Print a table of statistics for all tags in the database"""

    mrich.h1("hippo.calculate_scaffolds")

    mrich.h3("Params")
    mrich.var("database", database)

    animal = setup_animal(database=database, backup=False)
    animal.tags.summary()


def main():
    app()


if __name__ == "__main__":
    main()



import mcol

from .compound import Compound
from .recipe import Recipe

import mrich


class Reaction:
    """
    A :class:`.Reaction` is a simplified representation of a synthetic pathway to create a product :class:`.Compound`. Reactants (also :class:`.Compound` objects) as well as a reaction type are required.

    .. attention::

            :class:`.Reaction` objects should not be created directly. Instead use :meth:`.HIPPO.register_reaction` or :meth:`.HIPPO.reactions`

    """

    _table = "reaction"

    def __init__(
        self,
        db: "Database",
        id: int,
        type: str,
        product: int,
        product_yield: float,
    ) -> None:

        self._db = db
        self._id = id
        self._type = type
        self._product_id = product
        self._product = None
        self._product_yield = product_yield
        self._metadata = None

    ### PROPERTIES

    @property
    def id(self) -> int:
        """Returns the :class:`.Reaction` ID"""
        return self._id

    @property
    def type(self) -> str:
        """Returns the :class:`.Reaction` tyoe"""
        return self._type

    @property
    def product(self) -> "Compound":
        """Returns the reaction's product :class:`.Compound`"""
        if self._product is None:
            self._product = self.db.get_compound(id=self.product_id)
        return self._product

    @property
    def product_yield(self) -> float:
        """Returns the reaction's product yield (fraction)"""
        return self._product_yield

    @property
    def db(self) -> "Database":
        """Returns a pointer to the parent database"""
        return self._db

    @property
    def reactants(self) -> "CompoundSet":
        """Returns a :class:`.CompoundSet` of the reactants"""
        from .cset import CompoundSet

        return CompoundSet(self.db, indices=self.reactant_ids)

    @property
    def reaction_str(self) -> str:
        """Returns a string representing the reaction"""
        s = " + ".join([str(r) for r in self.reactants])
        s = f"{s} -> {str(self.product)}"
        return s

    @property
    def reactant_ids(self) -> set[int]:
        """Returns a set of reactant ID's"""
        return set(v for v in self.get_reactant_ids())

    @property
    def reactant_str_ids(self) -> str:
        """Return an SQL formatted tuple string of the reactant :class:`.Compound` IDs"""
        return str(tuple(self.reactant_ids)).replace(",)", ")")

    @property
    def product_id(self) -> int:
        """Returns the product :class:`.Compound` ID"""
        return self._product_id

    @property
    def product_smiles(self) -> str:
        """Product :class:`.Compound` SMILES string"""
        return self.product.smiles

    @property
    def reactant_smiles(self) -> list[str]:
        """List of reactant :class:`.Compound` SMILES strings"""
        return [r.smiles for r in self.reactants]

    @property
    def product_mol(self):
        """Product :class:`.Compound` ``rdkit.Chem.Mol`` object"""
        return self.product.mol

    @property
    def reactant_mols(self):
        """List of reactant :class:`.Compound` ``rdkit.Chem.Mol`` object"""
        return [r.mol for r in self.reactants]

    @property
    def price_estimate(self) -> float:
        """Estimate the price of this :class:`.Reaction`"""
        return self.db.get_reaction_price_estimate(reaction=self)

    @property
    def plain_repr(self) -> str:
        """Unformatted long string representation"""
        return f"{self}: {self.reaction_str} via {self.type}"

    @property
    def metadata(self) -> "MetaData":
        """Returns the compound's metadata dict"""
        if self._metadata is None:
            self._metadata = self.db.get_metadata(table="reaction", id=self.id)
        return self._metadata

    ### METHODS

    def get_reactant_amount_pairs(self, compound_object: bool = True) -> list[tuple]:
        """Returns pairs of reactants and their amounts

        :param compound_object: return :class:`.Compound` object instead of ID, (Default value = True)
        :returns: list of tuples containing :class:`.Compound` ID/object and amount in mg
        """

        compound_ids = self.db.select_where(
            query="reactant_compound, reactant_amount",
            table="reactant",
            key="reaction",
            value=self.id,
            multiple=True,
        )

        if compound_ids:
            if compound_object:
                return [
                    # (self.db.get_compound(id=id), amount/self.product_yield) for id, amount in compound_ids
                    (self.db.get_compound(id=id), amount)
                    for id, amount in compound_ids
                ]
            else:
                return compound_ids
        else:
            return []

    def get_reactant_ids(self) -> list[int]:
        """Returns list of reactants :class:`.Compound` IDs

        :returns: list of :class:`.Compound` IDs
        """

        compound_ids = self.db.select_where(
            query="reactant_compound",
            table="reactant",
            key="reaction",
            value=self.id,
            multiple=True,
        )

        if compound_ids:
            return [id for id, in compound_ids]
        else:
            return []

    def get_recipes(
        self,
        amount: float = 1,  # in mg
        debug: bool = False,
        pick_cheapest: bool = False,
        permitted_reactions: "None | ReactionSet" = None,
        supplier: str | None = None,
    ) -> "Recipe | list[Recipe]":
        """Get a :class:`.Recipe` describing how to make the product

        :param amount: Amount in ``mg``, defaults to ``1``
        :param debug: Increase verbosity, (Default value = False)
        :param pick_cheapest: pick the cheapest :class:`.Recipe`, (Default value = False)
        :param permitted_reactions: Limit the reactions to consider to members of this set, (Default value = None)
        :param supplier: Limit to reactants from this supplier (Default value = None)
        :returns: :class:`.Recipe` object or list thereof

        """

        from .recipe import Recipe

        return Recipe.from_reaction(
            self,
            amount=amount,
            debug=debug,
            pick_cheapest=pick_cheapest,
            permitted_reactions=permitted_reactions,
            supplier=supplier,
        )

    def summary(
        self,
        draw: bool = True,
    ) -> None:
        """Print a summary of this reaction's information

        :param draw: draw the reaction compounds (Default value = True)

        """

        print(f"id={self.id}")
        print(f"type={self.type}")
        print(f"product={self.product}")
        print(f"product_yield={self.product_yield}")

        reactants = self.get_reactant_amount_pairs()
        print(f"reactants={reactants}")

        print(f"price_estimate={self.price_estimate}")

        if draw:
            self.draw()

    def draw(self) -> None:
        """Draw the molecules involved in this reaction"""

        from molparse.rdkit import draw_grid

        reactants = self.reactants

        product = self.product

        mols = [r.mol for r in reactants]
        mols.append(product.mol)

        labels = [f"+ {r}" if i > 0 else f"{r}" for i, r in enumerate(reactants)]
        labels.append(f"-> {product}")

        drawing = draw_grid(mols, labels=labels, highlightAtomLists=None)
        display(drawing)

    def check_chemistry(
        self,
        debug: bool = False,
    ) -> bool:
        """Sanity check the chemistry of this reaction

        :param debug: increase verbosity (Default value = False)
        """
        from .chem import check_chemistry

        return check_chemistry(self.type, self.reactants, self.product, debug=debug)

    def check_reactant_availability(
        self,
        supplier: None | str = None,
        debug: bool = False,
    ) -> bool:
        """Check the availability of reactant compounds

        :param supplier: Limit to quotes from this supplier (Default value = None)
        :param debug: increase verbosity (Default value = False)

        """

        if debug:
            mrich.var("reaction", self.id)
            mrich.var("reactants", self.reactant_ids)
            mrich.var("supplier", supplier)

        if supplier is None:

            triples = self.db.execute(
                f"""
                SELECT reactant_compound, SUM(quote_id), SUM(reaction_id) FROM reactant 
                LEFT JOIN quote ON quote_compound = reactant_compound
                LEFT JOIN reaction ON reaction_product = reactant_compound 
                WHERE reactant_reaction = {self.id}
                GROUP BY reactant_compound
            """
            ).fetchall()

        else:

            triples = self.db.execute(
                f"""
                WITH filtered_quotes AS
                (
                    SELECT * FROM quote
                    WHERE quote_supplier = "{supplier}"
                )
                SELECT reactant_compound, SUM(quote_id), SUM(reaction_id) FROM reactant 
                LEFT JOIN filtered_quotes ON quote_compound = reactant_compound
                LEFT JOIN reaction ON reaction_product = reactant_compound 
                WHERE reactant_reaction = {self.id}
                GROUP BY reactant_compound
            """
            ).fetchall()

        for reactant_compound, has_quote, has_reaction in triples:

            if debug:
                mrich.debug(
                    f"{reactant_compound=}, {bool(has_quote)=}, {bool(has_reaction)=}"
                )

            if has_quote:
                if debug:
                    mrich.debug(f"reactant={reactant_compound} has quote")
                continue

            if has_reaction:
                if debug:
                    mrich.debug(f"reactant={reactant_compound} has reaction")
                continue

            if debug:
                mrich.warning(f"No quote or reaction for reactant={reactant_compound}")

            return False

        return True

    def get_dict(
        self,
        smiles: bool = True,
        mols: bool = True,
    ) -> dict[str]:
        """Returns a dictionary representing this :class:`.Reaction`

        :param smiles: include smiles string (Default value = True)
        :param mols: include ``rdkit.Chem.Mol`` (Default value = True)

        """

        serialisable_fields = ["id", "type", "product_id", "reactant_ids"]

        data = {}
        for key in serialisable_fields:
            data[key] = getattr(self, key)

        if smiles:
            data["product_smiles"] = self.product_smiles
            data["reactant_smiles"] = self.reactant_smiles

        if mols:
            data["product_mol"] = self.product_mol
            data["reactant_mols"] = self.reactant_mols

        return data

    def _delete(self) -> None:
        """Delete this reaction and any related reactants, routes, and components"""

        route_ids = self.db.select_where(
            query="component_route",
            table="component",
            key=f"component_ref = {self.id} AND component_type = 1",
            multiple=True,
        )

        route_ids = [r for r, in route_ids]
        route_str_ids = str(tuple(route_ids)).replace(",)", ")")

        self.db.delete_where(
            table="component", key=f"component_route IN {route_str_ids}"
        )

        self.db.delete_where(table="route", key=f"route_id IN {route_str_ids}")

        self.db.delete_where(table="reactant", key="reaction", value=self.id)

        self.db.delete_where(table="reaction", key="id", value=self.id)

    ### DUNDERS

    def __str__(self) -> str:
        """Unformatted string representation"""
        return f"R{self.id}"

    def __repr__(self) -> str:
        """ANSI Formatted string representation"""
        return f"{mcol.bold}{mcol.underline}{self.plain_repr}{mcol.unbold}{mcol.ununderline}"

    def __rich__(self) -> str:
        """Rich Formatted string representation"""
        return f"[bold underline]{self.plain_repr}"

    def __eq__(
        self,
        other: "int | Reaction",
    ) -> bool:
        """compare this reaction to a :class:`.Reaction` object or ID"""

        match other:
            case int():
                return self.id == other

            case Reaction():

                if self.type != other.type:
                    return False

                if self.product != other.product:
                    return False

                if self.reactant_ids != other.reactant_ids:
                    return False

                return True

            case _:
                raise NotImplementedError

    def __hash__(self) -> int:
        """Integer hash from ID"""
        return self.id



def reactions_from_row(*, animal, row, num_steps):

    reaction_ids = set()

    for step in range(num_steps):

        step += 1

        # get relevant fields
        reaction_name = row[f"{step}_reaction"]

        product = None
        reactants = []

        if reactant1_smiles := row[f"{step}_r1_smiles"]:
            reactants.append(animal.register_compound(smiles=reactant1_smiles))

        if reactant2_smiles := row[f"{step}_r2_smiles"]:
            reactants.append(animal.register_compound(smiles=reactant2_smiles))

        if product_smiles := row[f"{step}_product_smiles"]:
            product = animal.register_compound(smiles=product_smiles)

        assert product
        assert reactants

        reaction = animal.register_reaction(
            type=reaction_name,
            product=product,
            reactants=reactants,
        )

        assert reaction

        reaction_ids.add(reaction.id)

    return animal.reactions[reaction_ids]
